import os
import re
import subprocess

import astropy.io.fits as fits
from astropy import wcs
from astropy.modeling import models, fitting
import numpy as np
import scipy.ndimage as ndimage
import scipy.stats
from scipy.interpolate import interp1d
import time
from copy import copy

import multiprocessing as mp

#different imports depending on if python2.7 or python3
import sys
if sys.version_info < (3,0):
    #python 2.7 behavior
    import ConfigParser
else:
    import configparser as ConfigParser
    
from pyklip.instruments.Instrument import Data
from pyklip.instruments.utils.nair import nMathar

from pyklip.parallelized import high_pass_filter_imgs
from pyklip.fakes import gaussfit2d
from pyklip.fakes import gaussfit2dLSQ

class NIRC2Data(Data):
    """
    A sequence of Keck NIRC2 ADI Data. Each NIRC2Data object has the following fields and functions

    Args:
        filepaths: list of filepaths to files
        highpass: if True, run a Gaussian high pass filter (default size is sigma=imgsize/10)
                  can also be a number specifying FWHM of box in pixel units
        find_star: (default) 'auto' will first try to get the star center coordinates from the FITS
                  header PSFCENTX & PSFCENTY keywords, and if that fails it will do a Radon transform to
                  locate the star via the diffraction spikes (and store the star center in the header for
                  future use). True will force the Radon transform; False will skip the Radon transform
                  even if no center is found in the header.
        guess_star: (default) None. Otherwise an [x, y] coordinate guess for where the star is for find_star. 
                    The code will automatially guess the star location based on the FPM otherwise. 


    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        wcs: Array of N wcs astormetry headers for each image.
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        creator: string for creator of the data (used to identify pipelines that call pyklip)
        klipparams: a string that saves the most recent KLIP parameters

    Methods:
        readdata(): reread in the dadta
        savedata(): save a specified data in the GPI datacube format (in the 1st extension header)
        calibrate_output(): flux calibrate the output data
    """
    ##########################
    ###Class Initilization ###
    ##########################
    #some static variables to define the GPI instrument
    lenslet_scales = {}  # in arcsec/pixel
    centralwave = {}  # in microns
    fpm_diam = {}  # in pixels
    fpm_yx = {}  # in pixels
    flux_zeropt = {}
    pupil_diam = {}  # in meters
    #spot_ratio = {} #w.r.t. central star
    #ifs_rotation = 0.0  # degrees CCW from +x axis to zenith

    observatory_latitude = None

    ## read in GPI configuration file and set these static variables
    package_directory = os.path.dirname(os.path.abspath(__file__))
    configfile = package_directory + "/" + "NIRC2.ini"
    config = ConfigParser.ConfigParser()
    try:
        config.read(configfile)
        #get list of pixel scales for different cameras and time periods
        cameras = ['narrow_pre150413', 'narrow_post150413', 'medium', 'wide']
        for cam in cameras:
            lenslet_scales[cam] = float(config.get("pixel_scales", "pixel_scale_{0}".format(cam))) # arcsec/pix
        #get IFS rotation
        #ifs_rotation = float(config.get("instrument", "ifs_rotation")) #degrees
        #get some information specific to each band
        bands = ['J', 'H', 'K', 'Ks', 'Kp', 'Lp', 'Ms']
        for band in bands:
            centralwave[band] = float(config.get("instrument", "cen_wave_{0}".format(band)))
            flux_zeropt[band] = float(config.get("instrument", "zero_pt_flux_{0}".format(band)))
            #spot_ratio[band] = float(config.get("instrument", "APOD_{0}".format(band)))
        fpms = ['corona100', 'corona150', 'corona200', 'corona300', 'corona400',
                'corona600', 'corona800', 'corona1000', 'corona1500', 'corona2000']
        for fpm in fpms:
            fpm_diam[fpm] = float(config.get("instrument", "fpm_diam_{0}".format(fpm))) # arcsec
        for fpm in fpms:
            fpm_yx['narrow_' + fpm] = eval(config.get("instrument", "fpm_yx_narrow_{0}".format(fpm))) # (y,x) tuple [pixels]
            fpm_yx['medium_' + fpm] = eval(config.get("instrument", "fpm_yx_medium_{0}".format(fpm))) # (y,x) tuple [pixels]
            fpm_yx['wide_' + fpm] = eval(config.get("instrument", "fpm_yx_wide_{0}".format(fpm))) # (y,x) tuple [pixels]
        pupils = ['incircle', 'largehex', 'smallhex', 'open']
        for pupil in pupils:
            pupil_diam[pupil] = float(config.get("instrument", "pupil_diam_{0}".format(pupil))) # meters
        
        observatory_latitude = float(config.get("observatory", "observatory_lat"))
    except ConfigParser.Error as e:
        print("Error reading GPI configuration file: {0}".format(e.message))
        raise e

    ####################
    ### Constructors ###
    ####################
    def __init__(self, filepaths=None, highpass=False, find_star='auto', meas_star_flux=False, guess_star=None):
        """
        Initialization code for NIRC2Data

        Note:
            see class docstring for argument details
        """
        super(NIRC2Data, self).__init__()
        self._output = None

        self.flipx = True ### TODO: is this right?????
        
        if filepaths is None:
            self._input = None
            self._centers = None
            self._filenums = None
            self._filenames = None
            self._PAs = None
            self._wvs = None
            self._wcs = None
            self._IWA = None
            self.spot_flux = None
            self.star_flux = None
            self.contrast_scaling = None
            self.prihdrs = None
            self.exthdrs = None
            self.lenslet_scale = None
            self.pupil_diam = None
        else:
            self.readdata(filepaths, highpass=highpass, find_star=find_star, meas_star_flux=meas_star_flux, guess_star=guess_star)


    ################################
    ### Instance Required Fields ###
    ################################
    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval

    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval

    @property
    def filenums(self):
        return self._filenums
    @filenums.setter
    def filenums(self, newval):
        self._filenums = newval

    @property
    def filenames(self):
        return self._filenames
    @filenames.setter
    def filenames(self, newval):
        self._filenames = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval

    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval

    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval

    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval


    ###############
    ### Methods ###
    ###############
    def readdata(self, filepaths, highpass=False, find_star='auto', meas_star_flux=False, guess_star=None):
        """
        Method to open and read a list of NIRC2 data

        Args:
            filespaths: a list of filepaths
            highpass: if True, run a Gaussian high pass filter (default size is sigma=imgsize/10)
                  can also be a number specifying FWHM of box in pixel units
            find_star: (default) 'auto' will first try to get the star center coordinates from the FITS
                  header PSFCENTX & PSFCENTY keywords, and if that fails it will do a Radon transform to
                  locate the star via the diffraction spikes (and store the star center in the header for
                  future use). True will force the Radon transform; False will skip the Radon transform
                  even if no center is found in the header.
        guess_star: (default) None. Otherwise an [x, y] coordinate guess for where the star is for find_star. 
                    The code will automatially guess the star location based on the FPM otherwise. 

        Returns:
            Technically none. It saves things to fields of the NIRC2Data object. See object doc string
        """
        #check to see if user just inputted a single filename string
        if isinstance(filepaths, str):
            filepaths = [filepaths]

        #make some lists for quick appending
        data = []
        filenums = []
        filenames = []
        rot_angles = []
        wvs = []
        centers = []
        wcs_hdrs = []
        star_fluxes = []
        spot_fluxes = []
        prihdrs = []
        
        #Create a threadpool for high pass filter
        pool = None
        if highpass:
            pool = mp.Pool()

        #extract data from each file
        for index, filepath in enumerate(filepaths):
            cube, center, pa, wv, astr_hdrs, filt_band, fpm_band, pupil, star_flux, spot_flux, prihdr, exthdr, camera, obsdate =\
                _nirc2_process_file(filepath, highpass=highpass, find_star=find_star, meas_star_flux=meas_star_flux, pool=pool, guess_star=guess_star)

            data.append(cube)
            centers.append(center)
            star_fluxes.append(star_flux)
            spot_fluxes.append(spot_flux)
            rot_angles.append(pa)
            wvs.append(wv)
            filenums.append(np.ones(pa.shape[0]) * index)
            wcs_hdrs.append(astr_hdrs)
            prihdrs.append(prihdr)

            #filename = np.chararray(pa.shape[0])
            #filename[:] = filepath
            filenames.append([filepath for i in range(pa.shape[0])])
        
        # Close threadpool
        if highpass:
            pool.close()
            pool.join()
        
        #convert everything into numpy arrays
        #reshape arrays so that we collapse all the files together (i.e. don't care about distinguishing files)
        data = np.array(data)
        dims = data.shape
        data = data.reshape([dims[0] * dims[1], dims[2], dims[3]])
        filenums = np.array(filenums).reshape([dims[0] * dims[1]])
        filenames = np.array(filenames).reshape([dims[0] * dims[1]])
        rot_angles = np.array(rot_angles).reshape([dims[0] * dims[1]])  # want North Up
        wvs = np.array(wvs).reshape([dims[0] * dims[1]])
        wcs_hdrs = np.array(wcs_hdrs).reshape([dims[0] * dims[1]])
        centers = np.array(centers).reshape([dims[0] * dims[1], 2])
        star_fluxes = np.array(star_fluxes).reshape([dims[0] * dims[1]])
        spot_fluxes = np.array(spot_fluxes).reshape([dims[0] * dims[1]])
        
        #select correct pixel scale
        #narrow camera pixel scale changed after servicing on 2015-04-13
        date_2015_4_13 = time.strptime("2015-4-13", "%Y-%m-%d")
        if camera=='narrow':
            if obsdate < date_2015_4_13:
                lenslet_scale = NIRC2Data.lenslet_scales['narrow_pre150413']
            elif obsdate >= date_2015_4_13:
                lenslet_scale = NIRC2Data.lenslet_scales['narrow_post150413']
        else:
            lenslet_scale = NIRC2Data.lenslet_scales[camera]
        self.lenslet_scale = lenslet_scale
        
        #set these as the fields for the GPIData object
        self._input = data
        self._centers = centers
        self._filenums = filenums
        self._filenames = filenames
        self._PAs = rot_angles
        self._wvs = wvs
        self._wcs = [None] #wcs_hdrs
        self.spot_flux = spot_fluxes
        if fpm_band in NIRC2Data.fpm_diam:
            self._IWA = NIRC2Data.fpm_diam[fpm_band]/lenslet_scale/2.0
        else:
            # for unocculted data
            self._IWA = 0
        self.star_flux = star_fluxes
        self.contrast_scaling = 1./star_fluxes #GPIData.spot_ratio[ppm_band]/np.tile(np.mean(spot_fluxes.reshape(dims[0], dims[1]), axis=0), dims[0])
        self.prihdrs = prihdrs
        self.pupil_diam = NIRC2Data.pupil_diam[pupil]
        self.flipx = False
        
        #self.exthdrs = exthdrs

    def savedata(self, filepath, data, klipparams = None, filetype = None, zaxis = None, center=None, astr_hdr=None,
                 fakePlparams = None, more_keywords=None):
        """
        Save data in a GPI-like fashion. Aka, data and header are in the first extension header

        Note: In principle, the function only works inside klip_dataset(). In order to use it outside of klip_dataset,
            you need to define the following attribute:
                dataset.output_centers = dataset.centers

        Inputs:
            filepath: path to file to output
            data: 2D or 3D data to save
            klipparams: a string of klip parameters
            filetype: filetype of the object (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube")
            zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
            astr_hdr: wcs astrometry header (None for NIRC2)
            center: center of the image to be saved in the header as the keywords PSFCENTX and PSFCENTY in pixels.
                The first pixel has coordinates (0,0)
            fakePlparams: fake planet params
            more_keywords (dictionary) : a dictionary {key: value, key:value} of header keywords and values which will
                            written into the primary header

        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(header=self.prihdrs[0]))
        hdulist.append(fits.ImageHDU(data=data, name="Sci"))

        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        filenames = np.unique(self.filenames)
        nfiles = np.size(filenames)
        hdulist[0].header["DRPNFILE"] = nfiles
        for i, thispath in enumerate(filenames):
            thispath = thispath.replace("\\", '/')
            splited = thispath.split("/")
            fname = splited[-1]
#            matches = re.search('S20[0-9]{6}[SE][0-9]{4}', fname)
            filename = fname#matches.group(0)
            hdulist[0].header["FILE_{0}".format(i)] = filename

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = "pyKLIP"
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        #if self.creator is None:
        #    hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)
        #else:
        #    hdulist[0].header['CREATOR'] = self.creator
        #    hdulist[0].header.add_history("Reduced by {0}".self.creator)

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = pyklipver

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = klipparams
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        if fakePlparams is not None:
            hdulist[0].header['FAKPLPAR'] = fakePlparams
            hdulist[0].header.add_history("pyKLIP reduction with fake planet injection parameters {0}".format(fakePlparams))

        if filetype is not None:
            hdulist[0].header['FILETYPE'] = filetype

        if zaxis is not None:
            #Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                #write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = klmode

        #use the dataset astr hdr if none was passed in
        #if astr_hdr is None:
        #    print(self.wcs[0])
        #    astr_hdr = self.wcs[0]
        if astr_hdr is not None:
            #update astro header
            #I don't have a better way doing this so we'll just inject all the values by hand
            astroheader = astr_hdr.to_header()
            exthdr = hdulist[0].header
            exthdr['PC1_1'] = astroheader['PC1_1']
            exthdr['PC2_2'] = astroheader['PC2_2']
            try:
                exthdr['PC1_2'] = astroheader['PC1_2']
                exthdr['PC2_1'] = astroheader['PC2_1']
            except KeyError:
                exthdr['PC1_2'] = 0.0
                exthdr['PC2_1'] = 0.0
            #remove CD values as those are confusing
            exthdr.remove('CD1_1')
            exthdr.remove('CD1_2')
            exthdr.remove('CD2_1')
            exthdr.remove('CD2_2')
            exthdr['CDELT1'] = 1
            exthdr['CDELT2'] = 1

        #use the dataset center if none was passed in
        if center is None:
            center = self.output_centers[0]
        if center is not None:
            hdulist[0].header.update({'PSFCENTX':center[0],'PSFCENTY':center[1]})
            hdulist[0].header.update({'CRPIX1':center[0],'CRPIX2':center[1]})
            hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))

        # store extra keywords in header
        if more_keywords is not None:
            for hdr_key in more_keywords:
                hdulist[0].header[hdr_key] = more_keywords[hdr_key]

        try:
            hdulist.writeto(filepath, overwrite=True)
        except TypeError:
            hdulist.writeto(filepath, clobber=True)
        hdulist.close()

    def calibrate_data(self, units="contrast"):
        """
        Calibrates the flux of the output of PSF subtracted data.

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.
            units: currently only support "contrast" w.r.t central star

        Return:
            img: calibrated image of the same shape (this is the same object as the input!!!)
        """
        if units == "contrast":
            for i in range(self.output.shape[0]):
                self.output[i] *= self.contrast_scaling[:, None, None]

    def calibrate_output(self, img, spectral=False, units="contrast"):
        """
        Calibrates the flux of the output of PSF subtracted data.

        Assumes the broadband flux calibration is just multiplication by a single scalar number whereas spectral
        datacubes may have a separate calibration value for each wavelength

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.
            units: currently only support "contrast" w.r.t central star

        Return:
            img: calibrated image of the same shape (this is the same object as the input!!!)
        """
        if units == "contrast":
            if spectral:
                # spectral cube, each slice needs it's own calibration
                numwvs = img.shape[0]
                img *= self.contrast_scaling[:numwvs, None, None]
            else:
                # broadband image
                img *= np.nanmean(self.contrast_scaling)

        return img

######################
## Static Functions ##
######################

def _nirc2_process_file(filepath, highpass=False, find_star='auto', meas_star_flux=False, pool=None, guess_star=None):
    """
    Method to open and parse a NIRC2 file

    Args:
        filepath: the file to open
        highpass: if True, run a Gaussian high pass filter (default size is sigma=imgsize/10)
                  can also be a number specifying FWHM of box in pixel units
        find_star: (default) 'auto' will first try to get the star center coordinates from the FITS
                  header PSFCENTX & PSFCENTY keywords, and if that fails it will do a Radon transform to
                  locate the star via the diffraction spikes (and store the star center in the header for
                  future use). True will force the Radon transform; False will skip the Radon transform
                  even if no center is found in the header.
        meas_star_flux: if True, measures the stellar flux
        pool: optional to pass along a threadpool (mainly for highpass filtering multiple images)
        guess_star: (default) None. Otherwise an [x, y] coordinate guess for where the star is for find_star. 
                    The code will automatially guess the star location based on the FPM otherwise. 

    Returns: (using z as size of 3rd dimension, z=1 for NIRC2)
        cube: 3D data cube from the file. Shape is (z,256,256)
        center: array of shape (z,2) giving each datacube slice a [xcenter,ycenter] in that order
        parang: array of z of the parallactic angle of the target (same value just repeated z times)
        wvs: array of z of the wavelength of each datacube slice. (For NIRC2, wvs = [None])
        astr_hdrs: array of z of the WCS header for each datacube slice. (For NIRC2, wcs = [None])
        filt_band: the band (Ks, Lp, Ms) used in the data (string)
        fpm_band: For NIRC2, fpm_band = [None]
        ppm_band: For NIRC2, ppm_band = [None]
        spot_fluxes: For NIRC2, array of z containing 1.0 for each image
        prihdr: primary header of the FITS file
        exthdr: For NIRC2, None
    """
    print("Reading File: {0}".format(filepath))
    hdulist = fits.open(filepath, mode='update', ignore_missing_end=True) # some NIRC2 FITS files are missing END card
    try:
        #grab the data and headers
        cube = hdulist[0].data
        exthdr = None #hdulist[1].header
        prihdr = hdulist[0].header
        
        obsdate = time.strptime(prihdr['DATE-OBS'], "%Y-%m-%d") # observation date

        #get some instrument configuration from the primary header
        filt_band = prihdr['FILTER'].split('+')[0].strip()
        fpm_band = prihdr['SLITNAME']
        camera = prihdr['CAMNAME']
        pupil = prihdr['PMSNAME']
        rotmode = prihdr['ROTMODE'].lower()

        #for NIRC2, we only have broadband but want to keep the GPI array shape to make processing easier
        if prihdr['CURRINST'].strip() == 'NIRC2':
            wvs = [1.0]
            
            # Pull PA from header if already there, or try to calculate it from header.
            if 'ROTNORTH' in prihdr.keys():
                parang = prihdr['ROTNORTH']*np.ones(1)
            else:
                try:
                    parang = get_pa(hdulist, obsdate=obsdate, rotmode=rotmode, write_hdr=True)*np.ones(1)
                except:
                    parang = np.nan*np.ones(1)
            
            # manual guess for star centering?
            if guess_star is None:
                # automatic guess based on fpm 
                guess_star = NIRC2Data.fpm_yx[('_').join([camera, fpm_band])]

            if find_star is True:
                # Ignore header and use Radon transform to find star center.
                try:
                    center = [get_star(hdulist, ctr=guess_star, obsdate=obsdate,
                                    hp_size=0, im_smooth=0, sp_width=0, rad=100, radon_wdw=400, smooth=1,
                                    write_hdr=True, pool=pool, silent=True)]
                except:
                    center = [[np.nan, np.nan]]
            elif find_star is False:
                # Only try to get star center from headers (no Radon transform).
                try:
                    center = [[prihdr['PSFCENTX'], prihdr['PSFCENTY']]]
                except:
                    center = [[np.nan, np.nan]]
            elif find_star=='auto':
                # Pull star center from header if already there, or try to find it via Radon transform.
                # Radon assumes star is near center of FPM and may fail if otherwise.
                if 'PSFCENTX' and 'PSFCENTY' in prihdr.keys():
                    center = [[prihdr['PSFCENTX'], prihdr['PSFCENTY']]]
                else:
                    try:
                        center = [get_star(hdulist, ctr=guess_star, obsdate=obsdate,
                                        hp_size=0, im_smooth=0, sp_width=0, rad=100, radon_wdw=400, smooth=1,
                                        write_hdr=True, pool=pool, silent=True)]
                    except:
                        center = [[np.nan, np.nan]]
            else:
                raise ValueError("Unsupported value for find_star; only 'auto', True, or False are accepted.")
            
            # No cube flipping anymore!
            flipped_cube = cube

            star_flux = np.nan
            if meas_star_flux:
                star_flux = calc_starflux(flipped_cube, center)
            cube = flipped_cube.reshape([1, flipped_cube.shape[0], flipped_cube.shape[1]])  #maintain 3d-ness
            astr_hdrs = np.repeat(None, 1)
            spot_fluxes = [[1]] #not suported currently
    finally:
        pass
    
    #high pass filter
    highpassed = False
    if isinstance(highpass, bool):
        if highpass:
            cube = high_pass_filter_imgs(cube, pool=pool)
            highpassed = True
    else:
        # should be a number
        if isinstance(highpass, (float, int)):
            highpass = float(highpass)
            fourier_sigma_size = (cube.shape[1]/(highpass)) / (2*np.sqrt(2*np.log(2)))
            cube = high_pass_filter_imgs(cube, filtersize=fourier_sigma_size, pool=pool)
            highpassed = True

    return cube, center, parang, wvs, astr_hdrs, filt_band, fpm_band, pupil, star_flux, spot_fluxes, prihdr, exthdr, camera, obsdate

def calc_starflux(cube, center):
    """
    Fits a 2D Gaussian to an image to calculate the peak pixel value of
    the central star. The code assumes an unobscurated PSF.

    Args:
        cube: 2D image array. Shape is (256,256)
        center: star center in image in (x,y)

    Returns:
        Amplitude: Best fit amplitude of the 2D Gaussian.
    """

    dims = cube.shape
    y, x = np.meshgrid( np.arange(dims[0]), np.arange(dims[1]) )

    # Initializing Model. Fixing the rotation and the X, Y location of the star.
    g_init = models.Gaussian2D(cube.max(), x_mean=center[0][0], y_mean=center[0][1], x_stddev=5, y_stddev=5, \
        fixed={'x_mean':True,'y_mean':True,'theta':True})

    # Initializing Levenburg-Marquart Least-Squares fitting routine.
    fit_g = fitting.LevMarLSQFitter()

    # Fitting the amplitude, x_stddev and y_stddev
    g = fit_g(g_init, y, x, cube)

    return [[g.amplitude]]

def measure_star_flux(img, star_x, star_y):
    """
    Measure star peak fluxes using a Gaussian matched filter

    Args:
        img: 2D frame with unobscured, unsaturated PSF
        star_x, star_y: coordinates of the star
    Return:
        star_f: star flux
    """

    flux, fwhm, xfit, yfit = gaussfit2d(img, star_x, star_y, refinefit=False)
    if flux == np.inf: flux == np.nan
    print(flux, fwhm, xfit, yfit)

    return flux

def get_pa(hdulist, obsdate=None, rotmode=None, mean_PA=True, write_hdr=True):
    """
    Given a FITS data-header unit list (HDUList), returns the NIRC2 PA in [radians].
    PA is angle of detector relative to sky; ROTMODE is rotator tracking mode;
    PARANG is parallactic angle astrometric; INSTANGL is instrument angle;
    ROTPOSN is rotator physical position.
    Additional PA offset of -0.252 or -0.262 deg is applied for NIRC2 narrow cam
    depending on observation date.
    NOTE that the PA sign is flipped at the very end before output to conform to
    pyKLIP's rotation convention.
    
    Inputs:
        hdulist: a FITS HDUList (NOT a single HDU).
        obsdate: date of observation; will try to get from prihdr if not provided.
        rotmode: 'vertical angle' for ADI mode with PA rotating on detector, or
                 'position angle' for mode with PA orientation fixed on detector.
        mean_pa: if True (default), return the mean PA during the exposure.
                 If False, return the PA at the start of the exposure only.
                 Only applies to vertical angle mode.
        write_hdr: if True (default), writes keys to file header and saves them.
    """
    
    prihdr = hdulist[0].header
    
    # Date of NIRC2 servicing that changed PA offset.
    date_2015_4_13 = time.strptime("2015-4-13", "%Y-%m-%d")
    
    # If don't have it, get observation date (UT) from header and make a time object.
    if obsdate is None:
        obsdate = time.strptime(prihdr['DATE-OBS'], "%Y-%m-%d")
    
    # Additional offset to narrow cam PA not included in INSTANGL keyword.
    # This offset changed after instrument servicing on April 13, 2015.
    if prihdr['CAMNAME'].lower() == 'narrow':
        if obsdate < date_2015_4_13:
            zp_offset = -0.252 # [deg]; from Yelda et al. 2010
        elif obsdate >= date_2015_4_13:
            zp_offset = -0.262 # [deg]; from Service et al. 2016
    else:
        zp_offset = 0.
        print("WARNING: No PA offset applied.")
    
    if rotmode is None:
        global _last_rotmode
        try:
            rotmode = prihdr['ROTMODE'].strip().lower()
            _last_rotmode = rotmode
        except:
            rotmode = _last_rotmode.lower()
    rotposn = prihdr['ROTPOSN'] # [deg]
    instangl = prihdr['INSTANGL'] # [deg]

    if rotmode.lower() == 'vertical angle':
        parang = prihdr['PARANG'] # [deg]
        pa_deg = parang + rotposn - instangl + zp_offset # [deg]
    elif rotmode.lower() == 'position angle':
        pa_deg = rotposn - instangl + zp_offset # [deg]
    else:
        raise NotImplementedError
    
    if mean_PA and (rotmode.lower() == 'vertical angle'):
        # Get info for PA smearing calculation.
        epochobj = prihdr['DATE-OBS']
        name = prihdr['targname']
        expref = prihdr['itime']
        coaddref = prihdr['coadds']
        sampref =  prihdr['sampmode']
        msrref =   prihdr['MULTISAM']
        xdimref =  prihdr['naxis1']
        ydimref =  prihdr['naxis2']
        tel = prihdr['TELESCOP']
        dec = prihdr['DEC'] + prihdr['DECOFF']
        if tel.lower() == 'keck ii':
            tel = 'keck2' # just cleaning up str
        
        # Calculate total time of exposure (integration + readout).
        if sampref == 2: totexp = ( expref + 0.18*(xdimref/1024.)**2) * coaddref
        if sampref == 3: totexp = ( expref + (msrref-1)*0.18*(xdimref/1024.)**2)*coaddref
        
        tinteg = totexp # [seconds]
        totexp = totexp/3600. # [hours]
        
        # Get hour angle at start of exposure.
        tmpahinit = prihdr['HA'] # [deg]
        ahobs = 24.*tmpahinit/360. # [hours]
        
        # Estimate vertical position angle at each second of the exposure.
        vp = [] #fltarr(round(3600.*totexp))
        for j in range(0, int(round(3600.*totexp))-1):
            ahtmp = ahobs + (j*1.+0.001)/3600. # [hours]
            vp.append(par_angle(ahtmp, dec, NIRC2Data.observatory_latitude))
            if j == 0: vpref = vp[0]
        vp = np.array(vp)
        
        # Handle case where PA crosses 0 <--> 360.
        vp[vp < 0.] += 360.
        vp[vp > 360.] -= 360.
        if vpref < 0.: vpref += 360.
        if vpref > 360.: vpref -= 360.
        
        # Check that images near PA=0 are handled correctly.
        if any(vp > 350) & any(vp < 10):
            vp[vp > 350] -= 360
        
        vpmean = np.nanmean(vp)
        
        if (vpmean < 0) & (vpref > 350):
            vpmean += 360.
        
        pa_deg_mean = pa_deg + (vpmean - vpref)
    else:
        pa_deg_mean = pa_deg
        vpmean = np.nan
        vpref = np.nan
        totexp = np.nan
    
    if write_hdr:
        prihdr['TOTEXP'] = (totexp, 'Total exposure time [hours]')
        prihdr['PASTART'] = (pa_deg, "Position angle at exposure start [deg]")
        prihdr['PASMEAR'] = ((vpmean - vpref), "Exposure's weighted-mean PA minus PASTART [deg]")
        prihdr['ROTNORTH'] = (pa_deg_mean, "Mean PA of North during exposure [deg]")
        
        hdulist.flush(output_verify='ignore')
    
    # Flip sign to conform to pyKLIP rotation convention.
    return pa_deg_mean

def par_angle(HA, dec, lat):
    """
    Compute the parallactic angle, given hour angle (HA [hours]),
    declination (dec [deg]), and latitude (lat [deg]).  Returns
    parallactic angle in [deg].
    """
    HA_rad = np.radians(HA*15.) # [hours] -> [rad]
    dec_rad = np.radians(dec)   # [deg] -> [rad]
    lat_rad = np.radians(lat)   # [deg] -> [rad]

    parallang = -np.arctan2(-np.sin(HA_rad),  # [rad]
                          np.cos(dec_rad)*np.tan(lat_rad) - np.sin(dec_rad)*np.cos(HA_rad))
    
    return np.degrees(parallang) # [deg]

def get_star(hdulist, ctr, obsdate, hp_size=0, im_smooth=0, sp_width=0, spike_angles=None,
              r_mask='all', rad=100, rad_out=np.inf, radon_wdw=400, smooth=1, PAadj=0.,
              write_hdr=True, pool=None, silent=True):
    """
    Runs Radon transform star-finding algorithm on image and (by default) saves the results
    in the original FITS header.
    
    Inputs:
        hdulist: a FITS HDUList (NOT a single HDU).
        ctr: (y,x) coordinate pair estimate for star position for image [pix].
        obsdate: date of observation; will try to get from prihdr if not provided.
        hp_size: size of high-pass filter box (via Fourier transform) in [pix].
        im_smooth: sigma of smoothing Gaussian function in [pix].
        sp_width: width of diffraction spike mask in [pix]; default is 0 (no masking).
        spike_angles: list of discrete angles from the assumed star positions along 
            which the radon transform will sum intensity to search for the true star
            position (it picks the maximum sum). These should match the angles
            of the strongest diffraction spikes [deg].
        r_mask: 'all' to mask out circle around ctr coords; anything else to do no radial masking.
        rad: r_mask=='all' will mask out all r <= rad [pix].
        rad_out: r_mask=='all' will mask out all r > rad_out [pix].
        radon_window: half width of the radon sampling region; size_window = image.shape[0]/2 is suggested.
            m & M:  The sampling region will be (-M*size_window, -m*size_window)U(m*size_window, M*size_window).
        smooth: smooth the radon cost function; for one pixel, replace it by the
            average of its +/- smooth neighbours; default = 2.
        PAadj: optional angle by which to rotate diffraction spike angles in [radians].
        write_hdr: (default) True will write the Radon transform star center to the original
            FITS header in PSFCENTX & PSFCENTY keywords.
        pool: multiprocessing pool for highpass filtering and other parallel uses.
        silent: (default) True to suppress additional output to stdout.
    
    Outputs:
        Returns [X,Y] list of Radon transform star center. Default is to also write
        the star coordinates to PSFCENTX & PSFCENTY in original FITS header.
    """
    from pyklip.instruments.utils.radonCenter import searchCenter
    from scipy.ndimage.filters import median_filter, gaussian_filter
    
    if not silent: print("Finding star...")
    data = hdulist[0].data.copy()
    hdr = hdulist[0].header
    
    # median filter data to replace NaN's with median values of neighbors.
    wh = np.where(np.isnan(data))
    data_medfilt = median_filter(np.nan_to_num(data), size=9)
    # replace all NaN in f with median values of neighbors.
    data[wh] = data_medfilt[wh]
    
    data = np.ma.masked_invalid(data)
    
    # Highpass filtering.
    if hp_size != 0:
        # should be a number
        fourier_sigma_size = (data.shape[1]/float(hp_size)) / (2*np.sqrt(2*np.log(2)))
        data_filt = high_pass_filter_imgs(np.array([data]), filtersize=fourier_sigma_size, pool=pool)[0]
    else:
        data_filt = data
    
    # Image smoothing.
    if im_smooth != 0:
        data_filt = gaussian_filter(data_filt, im_smooth)
    
    # Build cartesian coordinate grid and radius map.
    yy, xx = np.mgrid[0:data_filt.shape[0]:1, 0:data_filt.shape[1]:1]
    radii = np.sqrt((yy - ctr[0])**2 + (xx - ctr[1])**2)
    
    # Additional offset to narrow cam PA not included in INSTANGL keyword.
    # This offset changed after instrument servicing on April 13, 2015.
    date_2015_4_13 = time.strptime("2015-4-13", "%Y-%m-%d")
    if hdr['CAMNAME'].lower()=='narrow':
        if obsdate < date_2015_4_13:
            zp_offset = -0.252 # [deg]; from Yelda et al. 2010
        elif obsdate >= date_2015_4_13:
            zp_offset = -0.262 # [deg]; from Service et al. 2016
    elif hdr['CAMNAME'].lower()=='wide':
        zp_offset = 0.
    
    # Select angles along which to perform radon based on rotator mode.
    # Position angle (pa) of telescope optics only, from header info.
    if hdr['ROTMODE'].lower() == 'vertical angle':
        pa_tele = hdr['INSTANGL'] - zp_offset - hdr['ROTPOSN'] # [deg]
    # Other case is either 'position angle' mode or unidentified.
    else:
        pa_tele = hdr['PARANG'] # [deg]
    if spike_angles is None:
        # Angles at which diffraction spikes occur in NIRC2 data [deg].
        spike_angles = pa_tele + 30.0 + np.arange(3)*60.0 + PAadj # [deg]
    # Boolean mask excluding everything except diffraction spikes.
    spikemask = ~ make_spikemask(data_filt, hdr, ctr, spike_angles, yy, xx, sp_width) # ~ is inverse boolean
    
    mask_total = spikemask.copy()
    
    # Optionally mask stellar halo with additional circular mask centered on ctr.
    # Masked regions are r <= rad and r > rad_out.
    if r_mask=='all':
        mask_total[(spikemask==False) & (radii <= rad)] = True
        mask_total[(spikemask==False) & (radii > rad_out)] = True
    else:
        pass
    
    data_masked = np.ma.array(data_filt, mask=mask_total)
    
    # Perform radon transform search for star.
    if not silent: print("Performing radon transform search...")
    (x_radon, y_radon) = searchCenter(data_masked.filled(0.), ctr[1], ctr[0],
                            size_window=radon_wdw, size_cost=7, m=0.2, M=0.8,
                            smooth=smooth, theta=spike_angles)

    if not silent: print("radon y,x = {0}, {1}".format(y_radon, x_radon))
    
    if write_hdr:
        # Update the original FITS header with new star coordinates.
        hdr['PSFCENTX'] = (x_radon, 'Star X numpy coord')
        hdr['PSFCENTY'] = (y_radon, 'Star Y numpy coord')
        
        hdulist.flush(output_verify='ignore')
        if not silent: print("Wrote new star coordinates to FITS header in PSFCENTX & PSFCENTY.")
    
    return [x_radon, y_radon]

def make_spikemask(data, hdr, ctr, spike_angles, yy, xx, width=31):
    """
    Construct diffraction spike mask from FITS header information.
    
    data: 2-D ndarray image to be masked (just to get size of array).
    hdr: FITS header for image constructing mask for.
    ctr: (y,x) coordinates for center of diffraction spike pattern
            (usually the estimated star location).
    spike_angles: position angles for diffraction spikes in image [radians].
    yy: mgrid or indices 2-D array of pixel y-coordinates.
    xx: mgrid or indices 2-D array of pixel x-coordinates.
    width: int or float width of spike mask in [pixels]; 0 for no mask.
    """
    
    mask = np.zeros(data.shape, dtype=np.bool)
    
    if (width > 0) & (not np.isnan(width)):
        yy_ctr = yy - ctr[0]
        xx_ctr = xx - ctr[1]
        
        for ang in np.radians(spike_angles):
            # Slope times x for each spike.
            mx = np.tan(ang)*xx_ctr
            # Mask with span 0.5*width above and below spike.
            band = np.abs(0.5*width/np.cos(ang))
            # Change spikemask to True anywhere inside mask.
            spikemask = (yy_ctr <= (mx + band)) & (yy_ctr >= (mx - band))
            # Replace False with True in mask wherever either spikemask or mask is True.
            mask = mask | spikemask
    else:
        mask = np.ones(data.shape, dtype=np.bool)
    
    return mask




#!/usr/bin/env python

from __future__ import division

import sys
import os
import warnings
import glob

import numpy as np
from numpy import ma
from scipy import interpolate

from astropy.io import fits
from astropy import units
from astropy.modeling import models, fitting

try:
    from photutils import aperture_photometry, CircularAperture
except:
    print("P1640: photutils not available; spot photometry will fail.")

#for handling different python versions
if sys.version_info < (3,0):
    import ConfigParser
else:
    import configparser as ConfigParser

class P1640params:
    num_spots = 4
    refchan = 26
    nchan = 32 # can be overridden by cube
    channels = np.arange(32, dtype=np.int) # can be overridden by cube
    wlsol = np.linspace(969, 1797, 32)*1e-3 # used for estimating scaling
    aperture_refchan = 3.5 # aperture size in the reference channel
P1640params.reflambda = P1640params.wlsol[P1640params.refchan]
P1640params.scale_factors = P1640params.wlsol/P1640params.reflambda

##################################################
# Grid operations
##################################################
def get_centered_grid(img_shape, center):
    """
    Return a coordinate grid shifted to the center
    """
    grid = np.indices(img_shape)
    centered_grid = np.array([grid[0]-center[0], grid[1]-center[1]])
    return centered_grid

def get_rotated_grid(img_shape, center, angle):
    """
    Rotate a coordinate grid by some angle around a center
    """
    centered_grid = get_centered_grid(img_shape, center)
    rad_grid = np.linalg.norm(centered_grid, axis=0)
    

##################################################
# Masks
##################################################
def make_mask_circle(img_shape, center, R):
    """
    Make a circular mask, where everything inside a radius R around the center 
    is False and outside is True
    Input:
        img_shape: the shape of the image in (row, col)
        center: the center of the mask, in (row, col)
        R: the radius of the circle
    Returns:
        mask: a masked array of shape img_shape
    """
    mask = np.zeros(shape=img_shape, dtype=np.int)
    centered_grid = get_centered_grid(img_shape, center)
    rad_grid = np.linalg.norm(centered_grid, axis=0)
    mask[rad_grid <= R] = 1
    return ~ma.make_mask(mask)

def make_mask_donut(img_shape, center, R0, R1):
    """
    Make a donut mask centered on 'center' where the inside of the donut is 
    False and the outside of the donut is True
    """
    inner_mask = make_mask_circle(img_shape, center, R0)
    outer_mask = make_mask_circle(img_shape, center, R1)
    donut_mask = (outer_mask != inner_mask)
    return donut_mask

def make_mask_half_img(img_shape, center, angle):
    """
    Mask half the image, cutting it through the center at an arbitrary angle.
    Angle is measured *counterclockwise* from vertical, and should be an 
    astropy units object, otherwise assume degrees. 
    Input:
        img_shape: shape of image in (row, col)
        center: the center of the mask in (row, col)
        angle: angle measured counterclockwise from vertical, default in deg
    Output:
        mask: masked_array with a plane running through point (center) at angle
            (angle)
    """
    if type(angle) != units.quantity.Quantity:
        angle = angle*units.degree
    mask = np.zeros(shape=img_shape, dtype=np.int)
    centered_grid = get_centered_grid(img_shape, center)
    rad_grid = np.linalg.norm(centered_grid, axis=0)
    ang_grid = np.arctan2(centered_grid[0], centered_grid[1]) * units.radian
    rot_grid = ang_grid - angle
    #dx = centered_grid[1] + rad_grid * np.cos(rot_grid)
    dx = rad_grid * np.cos(rot_grid)
    mask[dx>=0] = 1
    return ~ma.make_mask(mask)


def make_mask_bar(img_shape, center, angle, width):
    """
    Make a bar mask where all the pixels inside a bar through the center of
    the image within some width are 1 and everything outside is 0
    Inputs:
        img_shape: the shape of the image in (row, col)
        center: the center of the mask, in (row, col)
        angle: angle measured counterclockwise from vertical, default in deg
        width: with of bar in pixels
    Returns:
        mask: a masked array where the values inside the bar are False and 
            outside the bar are True
    """
    if type(angle) != units.quantity.Quantity:
        angle = angle*units.degree
    dcenter = np.array([(width/2) * np.sin(angle),
                        (width/2) * np.cos(angle)])
    mask1 = make_mask_half_img(img_shape, center+dcenter, angle)
    mask2 = make_mask_half_img(img_shape, center-dcenter, angle)
   
    # now mask the half-plane you don't want
    pmask = make_mask_half_img(img_shape, center, angle-90*units.degree)
    
    mask = ((mask1 != mask2) & pmask) #np.abs(mask2 - mask1)*pmask
    
    return mask

def make_mask_grid_spots(img_shape, centers, rotated_spots=False, nchan=P1640params.nchan):
    """
    Make a mask that shows only the grid spots
    Input:
        img_shape: the shape of the image to mask in (row, col)
        centers: (Nchan x 2) array of centers of the mask in (row, col)
        rotated_spots: [False] make mask for normal (False) or rotated (True) grid spots
        nchan: number of spectral channels in the cube
    Returns:
        masks: Nspot x Nchan cube of masks
    """
    if rotated_spots is True:
        angles = units.Quantity([21, 108, 194, 290], unit=units.degree)
        center_region = np.linspace(45, 85, 32) # centers mask regions

    else: # default values
        angles = units.Quantity([-25, 63, 155, 243], unit=units.degree)
        center_region = np.linspace(65, 110, 32) # centers mask regions
    ang_width = 40
    region_width = 20*P1640params.scale_factors
    inner_radii = center_region - region_width
    outer_radii = center_region + region_width
    
    donuts = [make_mask_donut(img_shape, i[0], i[1], i[2])
              for i in zip(centers, inner_radii, outer_radii)]
    #donut = make_mask_donut(img_shape, center, inner_radius, outer_radius)

    bars = [[make_mask_bar(img_shape, center, ang, ang_width) for center in centers]
            for ang in angles]

    masks = [[~(donuts[j] & bar[j]) for j in range(nchan)] for i,bar in enumerate(bars)]
    
    return masks


def make_mask_refined_grid_spots(img_shape, centers, spots, nchan=32):
    """
    Make a new set of masks that are centered on the interpolated 
    grid spot locations
    Input:
        img_shape: x- and y-dimensions of image
        centers: nchan x 2 array of star positions
        spots: num_spots  x nchan x 2 array of spot positions
    """
    drad = 10 # pixel half-width in radius
    dtheta = 20 # pixel full-width in theta
    masks = [[] for i in spots]
    for i,spot in enumerate(spots):
        for chan in range(nchan):
        # get spot radius
            rad = np.linalg.norm((spot[chan] - centers[chan]))
            theta = np.arctan2(spot[chan][0]-centers[chan][0],
                               spot[chan][1]-centers[chan][1])
            donut_mask = make_mask_donut(img_shape, centers[chan], 
                                         rad-drad, rad+drad)
            bar_mask = make_mask_bar(img_shape, centers[chan],
                                     theta*180/np.pi, dtheta)
            mask = ~(donut_mask & bar_mask)
            masks[i].append(mask)
    return masks
            

##################################################
# Fitting
##################################################
def guess_grid_spot_loc(img):
    """
    get max pixel as initial guess of location
    """
    spot_pos = np.unravel_index(np.argmax(img), dims=img.shape)
    return spot_pos

def fit_grid_spot(img, center, loc=None):
    """
    Fit spot with a 2-D Gaussian
    Inputs:
        img: 2-D masked_array to fit
        center: center of the image in (row, col) order
        loc (optional): initial position guess in (row, col) order
    """
    cleaned_img = (~img.mask)*img.data # image that is 0 outside mask
    if loc is None:
        loc = guess_grid_spot_loc(cleaned_img)

    x_stddev = 3#img[:,loc[1]].std()
    y_stddev = 3#img[loc[0],:].std()
    init_theta = np.arctan2(loc[0]-center[0], loc[1]-center[1])
    g_init = models.Gaussian2D(amplitude=img.max(), 
                               x_mean=loc[1], y_mean=loc[0],
                               x_stddev=x_stddev, y_stddev=y_stddev,
                               theta = init_theta)
    g_init.x_stddev.fixed = True
    g_init.y_stddev.fixed = True


    fit = fitting.LevMarLSQFitter()
    y,x = np.indices(img.shape)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        g = fit(g_init, x, y, img)

    return g, np.array([g.y_mean.value, g.x_mean.value])


def fit_poly(ind, dep, order):
    """
    Takes in an array of positions in row, col format and returns a polynomial
    that fits them to col = b + C*row, where C is a vector of coefficients
    Fitting is done by least squares
    Input:
        ind: dependent variable (prob channel number)
        dep: independent variable (prob x or y spot position)
    Returns:
        array of polynomial coefficients
    """
    ind = np.array(ind)
    dep = np.array(dep)
    x = np.array([ind**i for i in range(order+1)]).T
    y = dep
    a = np.dot(x.T, x)
    b = np.dot(x.T, y)
    coeffs = np.dot(np.linalg.inv(a), b)
    return coeffs

def get_points_from_poly(ind, coeffs):
    """
    Get a polynomials coefficients and return the y-values given the independent values
    """
    ind = np.array(ind)
    x = np.array([ind**i for i in range(coeffs.size)]).T
    y = np.dot(x, coeffs)
    return y


def check_bad_spots(spot, centers):
    """
    Input:
        spot: y,x positions for a spot
        centers: y,x positions for the center
    Output: 
        fixed_spot: a fixed spot position y,x
    """
    fixed_spot = spot.copy()
    centered_spot = spot-centers
    rad_spot = np.linalg.norm(centered_spot, axis=1)
    bad_spots = np.where(rad_spot[1:]<rad_spot[:-1])[0]+1
    for bad in bad_spots:
        good_channels = np.concatenate((P1640params.channels[:bad-1], 
                                        P1640params.channels[bad+1:]))
        # fit them with a chosen polynomial (cubic for now)
        good_rad = np.linalg.norm(centered_spot[good_channels], axis=1)
        good_ang = np.arctan2(centered_spot[good_channels][:,0],
                              centered_spot[good_channels][:,1]).mean()
        rad_line = fit_poly(good_channels, good_rad, 3)
        new_rad = get_points_from_poly(P1640params.channels, rad_line)
        newyx = (new_rad*np.sin(good_ang), new_rad*np.cos(good_ang)) + centers.T
        residuals = abs(rad_spot - new_rad)
        worse = bad if residuals[bad] > residuals[bad-1] else bad-1
        fixed_spot[worse,0] = newyx[0][worse]
        fixed_spot[worse,1] = newyx[1][worse]   
    return fixed_spot

def check_bad_channels(rad_spot):
    """
    Check that the spot positions increase monotonically in radius.
    If they don't, return the positions that do not follow monotonically.
    Input:
        rad_spot: Nchan array of radial sep of a spot from star 
    Output: 
        bad_chans: a list of channel pairs that fail the check and need fixing
    """
    try:
        assert rad_spot.ndim==1
    except AssertionError:
        print("check_bad_channels received array with wrong dimensions, exiting")
        sys.exit()
    # get bad channels, in ambiguous pairs
    bad_chans0 = np.where(rad_spot[1:] < rad_spot[:-1])[0]
    bad_chans1 = bad_chans0+1
    # convert to list and sort
    bad_chans = list(zip(bad_chans0, bad_chans1))
    temp3 = list(bad_chans)
    return bad_chans

def fix_bad_channels(spot, centers, bad_chans):
    """
    Two cases: 
        1) a spot jumps inwards
        2) a spot jumps outwards
    In either case, remove both the failing spot and the one before it
    and fit a cubic to the remaining points. Then, fix the spot that is
    further from the fit.
    Input: 
       spot: Nchan x 2 array of positions for one spot
       centers: y,x positions for the star in each channel
       bad_chan: index of a spot that does not monotonically increase in radius
    Output:
       fixed_spot: Nchan x 2 array of fixed positions for the spot
    """
    fixed_spot = np.copy(spot)
    centered_spot = spot - centers
    rad_spot = np.linalg.norm(centered_spot, axis=1)
    channels = range(spot.shape[0])

    # remove all bad channels from fitting and interpolation channels
    good_channels = list(channels[:])
    try:
        for i in np.unique(np.ravel(bad_chans)): good_channels.remove(i)
    except ValueError:
        pass
    '''
    # separate procedures for middle channel vs end channels
    # check if 1 is a bad channel (the lowest possible bad channel)
    if 1 in np.ravel(bad_chans): 
        # use the four lowest good channels to fit a quadratic
        fit_channels = good_channels[:5]
        #good_channels = channels[bad_chan+1:bad_chan+5]
        fit_rad = rad_spot[fit_channels] 
        rad_line = fit_poly(fit_channels, fit_rad, 2)
        new_rad = get_points_from_poly(channels, rad_line)
    # check if last channel is a bad channel
    elif channels[-1] in np.ravel(bad_chans): 
        # use the four highest good channels to fit a quadratic
        fit_channels = good_channels[-4:]
        fit_rad = rad_spot[fit_channels]
        rad_line = fit_poly(fit_channels, fit_rad, 2)
        new_rad = get_points_from_poly(channels, rad_line)
    # if bad channel is in the middle, we can interpolate 
    else:  # interpolation NOPE fit a cubic to all the good points
        # remove two channels from the good set
        good_rad = rad_spot[good_channels]
        rad_line = fit_poly(good_channels, good_rad, 3)
        new_rad = get_points_from_poly(channels, rad_line)
        #cubeinterp = interpolate.interp1d(good_channels, good_rad, kind='cubic')
        # splice in the interpolated values
        #new_rad = good_rad.tolist()
        #for i in np.ravel(bad_chans):
        #    new_rad.insert(i, cubeinterp(i))
        #new_rad = np.array(new_rad)
    '''
    good_rad = rad_spot[good_channels]
    rad_line = fit_poly(good_channels, good_rad, 3)
    new_rad = get_points_from_poly(P1640params.channels, rad_line)
    # to get x and y, just use the average angle
    good_ang = np.arctan2(centered_spot[good_channels][:,0],
                          centered_spot[good_channels][:,1]).mean()
    newyx = (new_rad*np.sin(good_ang), new_rad*np.cos(good_ang)) + centers.T

    # test bad channel pairs against fit
    residuals = abs(rad_spot - new_rad)
    for bad_chan in bad_chans:
        worse = bad_chan[0] if residuals[bad_chan[0]] > residuals[bad_chan[1]] else bad_chan[1]
        fixed_spot[worse,0] = newyx[0][worse]
        fixed_spot[worse,1] = newyx[1][worse]   

    return fixed_spot

##################################################
# Write spots to memory or disk
##################################################

def write_spots_to_header(spots, fitsfile):
    """
    Write the spot positions to a fits header
    Input:
        spots: 4 x Nchan x 2 array
        fitsfile: full path to a fits file whose header you want to modify
    """
    hdulist = fits.open(fitsfile)
    hdu = hdulist[0]
    header = hdu.header
    header['Spot0x'] = spots[0,:,1]
    header['Spot0y'] = spots[0,:,0]
    header['Spot1x'] = spots[1,:,1]
    header['Spot1y'] = spots[1,:,0]
    header['Spot2x'] = spots[2,:,1]
    header['Spot2y'] = spots[2,:,0]
    header['Spot3x'] = spots[3,:,1]
    header['Spot3y'] = spots[3,:,0]
    return hdu

def write_spots_to_file(data_filepath, spot_positions, output_dir=None,
                        spotid=None, ext=None, overwrite=True):
    """
    Write one file for each spot to the directory defined at the top of
    this file. Output file name is data_filename -fits +spoti.csv.
    Format is (row, col). Will overwrite existing files.
    Input:
        data_filename: the base name of the file with the spots
        spot_positions: Nspot x Nchan x 2 array of spot positions
        output_dir: directory to write the output files
        overwrite: (True) overwrite existing spot files
        spotid: (-spoti) identifier for the 4 different spot files
        ext: (csv) file extension
    Returns:
        None
        writes a file to the output dir whose name corresponds to the cube 
        used to generate the spots + spotidN.ext (N is 0-3)
    """
    data_filename = os.path.basename(data_filepath)
    exists = glob.glob(os.path.join(output_dir,data_filename)+"*")

    # If you shouldn't overwrite existing files, quit here
    if (exists) and (not overwrite):
        print("Spot files exist and overwrite is False, skipping...")
        return

    # if output_dir, spotid, and ext are NOT specified, used P1640.ini as defaults
    if np.any([i is None for i in [output_dir, spotid, ext]]):
        # default config file
        config = ConfigParser.ConfigParser()
        config.read("/data/home/jaguilar/pyklip/pyklip/instruments/P1640.ini")
        if output_dir is None:
            output_dir = config.get("spots", "spot_file_path")
            print("Using value in P1640.ini for spot output directory: " + output_dir)
        if spotid is None:
            spotid = config.get("spots", "spot_file_postfix")
            print("Using value in P1640.ini for spot file ID: " + spotid)
        if ext is None:
            ext = config.get("spots", "spot_file_ext")
            print("Using value in P1640.ini for spot file ext: " + ext)
    
    try:
        for i, spot in enumerate(spot_positions):
            data_filename = os.path.basename(data_filepath)
            spotnum = spotid+'{0}'.format(i)
            output_filename = os.path.splitext(data_filename)[0]+spotnum+'.'+ext
            output_filepath = os.path.join(output_dir, output_filename)
            np.savetxt(output_filepath, spot, delimiter=",",
                           header='row,column')
            print(os.path.basename(output_filepath) + " written")
    except:
        # implement error handling later?
        pass
    return
        

##################################################
# Complete spot extraction
##################################################

def get_initial_spot_guesses(cube, rotated_spots=False):
    """
    """
    nchan = cube.shape[0]
    channels = np.arange(nchan, dtype=np.int)
    img_shape = cube.shape[1:]
    init_centers = img_shape*np.ones((nchan, 2))/2

    spot_masks = make_mask_grid_spots(img_shape, init_centers, rotated_spots=rotated_spots)

    spot_locs = np.zeros((P1640params.num_spots, nchan, 2))
    masked_cubes =  ma.masked_array([ma.masked_array(cube, mask=i) 
                                     for i in spot_masks])
   # get reference channel grid spots:
    for i in range(P1640params.num_spots):
        img = masked_cubes[i, P1640params.refchan]
        g, spot_locs[i,P1640params.refchan] = fit_grid_spot(img, init_centers[P1640params.refchan])

    # get initial guesses for the rest
    centered_spots = spot_locs[:,P1640params.refchan] - init_centers[P1640params.refchan]
    init_rad = np.linalg.norm(centered_spots, axis=1)
    init_theta = np.arctan2(centered_spots[:,0], centered_spots[:,1])
    init_spots = np.rollaxis(np.rollaxis(
        np.array([np.outer(init_rad, P1640params.scale_factors).T*np.sin(init_theta),
                  np.outer(init_rad, P1640params.scale_factors).T*np.cos(init_theta)]),
        2, 0), 2, 1) + init_centers

    return init_spots

def fit_grid_spots(masked_cubes, centers, spots_guesses):
    """
    Wrapper for fit_grid_spot to loop over all four spots
    Input:
        masked_cubes: Nspot x Nchan x Npix x Npix array of masks
        centers: Nchan x 2 array of (row, col) guesses for spot centers
        spot_guesses: Nspot x Nchan x 2 (row, col) guesses for spot locations
    Output:
        spot_fits: List of astropy.model fits
        spot_locs: Nspot x Nchan x 2 array of spot locations from fitting
    """
    nchan = masked_cubes.shape[1]
    spot_locs = np.zeros((P1640params.num_spots, nchan, 2))
    spot_fits = [[] for i in range(P1640params.num_spots)]
    for i in range(P1640params.num_spots):
        images = masked_cubes[i] 
        for chan in range(nchan):
            g, spot_locs[i,chan] = fit_grid_spot(images[chan], centers[chan], 
                                                     loc=spots_guesses[i,chan])
            spot_fits[i].append(g)
    return spot_fits, spot_locs



#################################
# Single-cube methods
#################################

def get_single_cube_spot_positions(cube, rotated_spots=False):
    """
    Return the spot positions for a single cube
    Input:
        cube: a data cube from P1640
        rotated_spots: (False) if True, use the rotated masks
    Output:
        spot_array: Nspots x Nchan x 2 array of spot positions. 
    """
    #################################
    # some unavoidable initializations
    nchan = cube.shape[0]
    channels = np.arange(nchan, dtype=np.int)
    img_shape = cube.shape[1:]
    init_centers = img_shape*np.ones((nchan, 2))/2
    spot_masks = make_mask_grid_spots(img_shape, init_centers, rotated_spots=rotated_spots)
    masked_cubes =  ma.masked_array([ma.masked_array(cube, mask=i) 
                                     for i in spot_masks])
    #################################

    #################################
    # Initial pass
    init_spots = get_initial_spot_guesses(cube, rotated_spots)

    # now, fit rest of spots using initial guesses
    spot_fits, spot_locs = fit_grid_spots(masked_cubes, init_centers, init_spots)
    #################################
    # Calculate centers
    # At each channel, fit lines through opposing spots
    centers = get_single_cube_star_positions(spot_locs)
    # Fix 'bad' spots:
    # Fit the radial separation and get x and y from that
    fixed_spot_locs = np.copy(spot_locs)
    centered_spots = fixed_spot_locs - centers
    for i in range(P1640params.num_spots):
        rad_spots = np.linalg.norm(fixed_spot_locs[i] - centers,
                                   axis=-1)
        bad_channels = check_bad_channels(rad_spots)
        while len(list(bad_channels)) != 0:
            fixed_spot_locs[i] = fix_bad_channels(fixed_spot_locs[i],
                                                  centers, bad_channels)
            # update radial spot distances to check they've all been corrected
            rad_spots = np.linalg.norm(fixed_spot_locs[i] - centers, 
                                       axis=-1)
            bad_channels = check_bad_channels(rad_spots)

    # update centers using new positions
    centers = get_single_cube_star_positions(fixed_spot_locs)

    #################################
    # Second pass, with new masks
    refined_masks = make_mask_refined_grid_spots(img_shape, centers, 
                                                 fixed_spot_locs)
    refined_masked_cubes =  ma.masked_array([ma.masked_array(cube, mask=i) 
                                             for i in refined_masks])
    # now, fit spots and again check for bad spots
    refined_spot_locs = np.zeros(spot_locs.shape)
    for i in range(P1640params.num_spots):
        images = refined_masked_cubes[i]
        for chan in range(nchan):
            g, refined_spot_locs[i, chan] = fit_grid_spot(images[chan], 
                                                          centers[chan], 
                                                          loc=fixed_spot_locs[i,chan])
            spot_fits[i][chan] = g
    # check bad spots again
    for i in range(P1640params.num_spots):
        rad_spots = np.linalg.norm(refined_spot_locs[i] - centers, 
                                   axis=-1)
        bad_channels = check_bad_channels(rad_spots)
        while len(list(bad_channels)) != 0:
            refined_spot_locs[i] = fix_bad_channels(refined_spot_locs[i],
                                                    centers, bad_channels)
            # update radial spot distances to check they've all been corrected
            rad_spots = np.linalg.norm(refined_spot_locs[i] - centers,
                                       axis=-1)
            bad_channels = check_bad_channels(rad_spots)
    '''
    for i in range(P1640params.num_spots):
        while np.any(np.where(rad_spots[i,1:]<rad_spots[i,:-1])[0]+1):
            refined_spot_locs[i] = check_bad_spots(spot_locs[i], centers)
            # update radial spot distances to check they've all been corrected
            centered_spots[i] = refined_spot_locs[i] - centers
            rad_spots[i] = np.linalg.norm(centered_spots[i], axis=-1)
    '''

    return refined_spot_locs

def get_single_file_spot_positions(fitsfile, rotated_spots=False):
    """
    Wrapper for get_single_cube_spot_positions
    """
    hdulist = fits.open(fitsfile)
    cube = hdulist[0].data
    spot_positions = get_single_cube_spot_positions(cube, rotated_spots)
    hdulist.close()
    return spot_positions


#################################
# Operations on spot arrays
#################################

def get_single_cube_star_positions(spot_array):
    """
    Using the spot positions for a single cube, find the star position at each wavelength.
    Input:
        spot_array: Nspots x Nchan x 2 array of (row, column) spot positions
    Output:
        star_array: Nchan x 2 array of (row, column) star positions
    """
    nspots  = P1640params.num_spots # always 4, because physics 
    nchan = spot_array.shape[-2]

    star_position = np.zeros((nchan, 2))

    for chan in range(nchan):
        line20 = fit_poly(np.array((spot_array[0,chan,1], 
                                    spot_array[2,chan,1])),
                          np.array((spot_array[0,chan,0], 
                                    spot_array[2,chan,0])),
                          order=1)
        line31 = fit_poly(np.array((spot_array[1,chan,1], 
                                    spot_array[3,chan,1])),
                          np.array((spot_array[1,chan,0], 
                                    spot_array[3,chan,0])),
                          order=1)
        x_star = (line20[0]-line31[0])/(line31[1]-line20[1])
        y_star = (line20[1]*line31[0]-line31[1]*line20[0])/(line20[1]-line31[1])
        star_position[chan] = (y_star, x_star)
    
    return star_position
   
def get_single_cube_scaling_factors(spot_array, star_array=None):
    """
    Get the scaling factors for a single cube
    Input:
        spot_array: Nspots x Nchan x 2 array of (row, column) spot positions
        star_array: Nchan x 2 array of (row, column) star positions.
            If not supplied, spot_array will be calculated from spot_array
    Output:
        scaling: Nspots x Nchan array of scaling factors, normalized to
            P1640params.refchan
    """
    if star_array is None:
        star_array = get_single_cube_star_positions(spot_array)
    centered_spots = spot_array - star_array
    rad_spots = np.linalg.norm(centered_spots, axis=-1)
    scaling = rad_spots / rad_spots[:, P1640params.refchan][:,None] 
    #scaling = rad_spots[:, P1640params.refchan][:,None] / rad_spots
    return scaling

def get_single_file_scaling_and_centering(fitsfile):
    """
    Take a single fits file, and return the star positions and 
    scaling factors
    See also: get_scalign_and_centering_from_spots
    Input:
        fitsfile: a single fits file with a P1640 cube
    Output:
        scaling_factors: scaling factors for each slice of the cube
        star_positions: star positions in each slice of the cube
    """
    hdulist = fits.open(fitsfile)
    cube = hdulist[0].data
    spot_positions = get_single_cube_spot_positions(cube)
    star_positions = get_single_cube_star_positions(spot_positions)
    scaling_factors = get_single_cube_scaling_factors(spot_positions, star_positions)
    hdulist.close()
    return scaling_factors, star_positions



##################################
# Spot photometry
##################################

def get_single_cube_spot_photometry(cube, spot_positions):
    """
    Do aperture photometry on the spots. Will need to be careful about
    aperture size for future comparison
    Input:
        cube: Nchan x Npix x Npix data cube to do photometry
        spot_positions: Nspot x Nchan x 2 spot positions for apertures
        scaling_factors: Nchan array for scaling apertures with wavelength
    Output:
        spot_phot: Nspot x Nchan array of spot photometry and spot errors
    """
    nchan = cube.shape[0]
    apsize = P1640params.aperture_refchan
    photometry = []
    # need scaling factors to scale aperture
    scaling_factors = get_single_cube_scaling_factors(spot_positions).mean(axis=0)
    for chan in range(nchan):
        positions = [i[::-1] for i in spot_positions[:,chan,:]] # need to reverse row, col
        apertures = CircularAperture(positions = positions,
                                     r = apsize*scaling_factors[chan])
        photometry.append(aperture_photometry(cube[chan], apertures))
    spot_phot = np.array([i['aperture_sum'] for i in photometry]).T
    return spot_phot

def get_single_cube_spot_positions_and_photometry(cube):
    """
    Wrapper that combines get_single_cube_spot_positions and
    get_single_cube_spot_photometry
    Input:
        cube: a datacube in P1640 format
    Output:
        spot_positions: Nspots x Nchan x 2 array of spot positions. 
        spot_photometry: Nspots x Nchan x 1 array of spot fluxes
    """
    spot_positions = get_single_cube_spot_positions(cube)
    spot_photometry = get_single_cube_spot_photometry(cube, spot_positions)
    return spot_positions, spot_photometry


#################################
# Multi-cube wrappers
#################################

def get_spot_positions(fitsfiles):
    """
    Return the spot positions for a set of data cubes. Really just a wrapper
    for get_single_cube_spot_positions
    Input:
        fitsfiles: a list of P1640 fits files
    Output:
        spot_array: Nfile x 4 x Nchan x 2 array of spot positions
    """
    # support providing a single file
    if isinstance(fitsfiles, str):
        fitsfiles = [fitsfiles]

    spot_array  = np.array([get_single_cube_spot_positions(fits.getdata(f))
                           for f in fitsfiles])
    return spot_array

def get_star_positions(spot_array):
    """
    Get the center of a set of 4 spots for a single cube
    Input:
        spot_locations: Nspot x Nlambda x 2 array of [row, col] spot positions
    Returns:
        star_array: Nlambda x 2 array of [row, col] star positions
    """
    num_spots=4
    
    # find the pairs of opposite spots
    # sorry for the confusing python shorthand
    spot_list = list(range(num_spots))
    pairs=[]
    pair1 = 0
    pairs.append([pair1, np.linalg.norm(spot_array[pair1] - spot_array, axis=-1).mean(axis=-1).argmax()])
    for i in sorted(pairs[0])[::-1]:
        spot_list.pop(i)
    pair2 = spot_list.pop(0)
    pairs.append([pair2, np.linalg.norm(spot_array[pair2] - spot_array, axis=-1).mean(axis=-1).argmax()])

    # Re-order the array so that channels are on the first axis
    if spot_array.shape[-2] != num_spots:
        spotaxid = np.where(np.array(spot_array.shape) == num_spots)[0][0]
        spot_array = np.rollaxis(spot_array, spotaxid,-1)
    orig_shape = spot_array.shape
    orig_ndim = spot_array.ndim
    if orig_ndim == 4: # more than one cube - 
        # collapse channels since we'll be processing the slices independently
        spot_array = spot_array.reshape((orig_shape[-4]*orig_shape[-3], 
                                         orig_shape[-2], orig_shape[-1]))
    
    # each element of spot_array is a set of four co-wavelength spots to be centered
    nslices = len(spot_array)
    star_positions = np.zeros((len(spot_array), 2))
    for chan in range(nslices):
        b20,m20 = fit_poly(spot_array[chan,pairs[0],1], spot_array[chan,pairs[0],0], order=1)
        b31,m31 = fit_poly(spot_array[chan,pairs[1],1], spot_array[chan,pairs[1],0], order=1)
        x_star = (b31-b20)/(m20-m31)  
        y_star = (m31*b20-m20*b31)/(m31-m20)
        star_positions[chan] = (y_star, x_star)
    return star_positions
    # for the final shape, separate out the cubes again
    new_shape = list(orig_shape)
    new_shape.pop(-2) # remove the Nspots index
    star_positions = star_positions.reshape(new_shape)
    return star_positions


def get_spot_positions_and_photometry(fitsfiles):
    """
    Wrapper that combines get_single_cube_spot_positions and 
    get_single_cube_spot_photometry
    Accept a list of fits files and returns the spot positions and
    spot photometry
    Input:
        fitsfiles: a list of P1640 fits files
    Output: 
        spot_array: Nfiles x Nspots x Nchan x 2 array of (row, col) positions
        spot_phot: Nfils x Nspots x Nchan array of spot photometry
    """
    if isinstance(fitsfiles, str):
            filepaths = [fitsfiles]

    spot_positions = []
    photometry = []
    for f in fitsfiles:
        hdulist = fits.open(f)
        cube = hdustlst[0].data
        spot_positions.append(get_single_cube_spot_positions(cube))
        photometry.append(get_single_cube_spot_photometry(cube, spot_positions))
        hdulist.close()
    return np.array(spot_positions, spot_photometry)

def get_scaling(spot_array, star_array=None, return_mean=True):
    """
    Wrapper for get_single_cube_scaling_factors, to handle multiple cubes
    Input:
        spot_array: (Nfiles) x Nspots x Nchan x 2 array in (row, col) order
        star_array: (Nfiles) x Nchan x 2 array of (row, column) star positions.
            If not supplied, spot_array will be calculated from spot_array
        return_mean: (default False) If true, return mean scaling factor for
            each channel (useful for multiple cubes)
    Output:
        scaling_array: (Nfiles) x Nchan array of scaling fators
    """
    if spot_array.ndim == 3:
        spot_array = np.expand_dims(spot_array, 0)
    if star_array is None:
        star_array = np.array([None for i in spot_array])
    scaling = []
    for spots, star in zip(spot_array, star_array):
        scaling.append(get_single_cube_scaling_factors(spots, star))
    scaling = np.array(scaling)

    if return_mean == True:
        scaling = np.mean(scaling, axis=-2)

    return scaling

def get_scaling_and_centering_from_spots(spot_positions, mean_scaling=True):
    """
    Accepts an array of spots, and returns the scaling factors and centers.
    See also: get_scaling_and_centering_from_files
    Input:
        spot_positions: Ncube x Nspot x Nchan x 2 array of (row, col) spot positions
        mean_scaling: [True] return the average scaling of the 4 spots
    Output:
        scaling_factors: Ncube x  Nchan array of scaling factors
        star_positions: Ncube x Nchan x 2 array of (row, col) star positions
    """
    star_positions = get_star_positions(spot_positions)
    scaling_factors = get_scaling(spot_positions, star_positions, mean_scaling)
    return scaling_factors, star_positions
    
def get_scaling_and_centering_from_files(files, mean_scaling=True):
    """
    Take some csv spot files, and return the star positions and 
    scaling factors for each datacube
    Wrapper for get_scaling_and_centering_from_spots
    Input:
        files: a list of fits files with data cubes
               if the files end in fits or csv, call appropriate routines
        mean_scaling: [True] return the mean scaling of the 4 spots

    Output:
        scaling_factors: scaling factors for each slice of each cube
        star_positions: star positions in each slice of each cube
    """
    spot_positions = [np.genfromtxt(f, delimiter=',') for f in files]
    spot_positions = np.array(spot_positions)
    scaling_factors, star_positions = get_scaling_and_centering_from_spots(spot_positions, mean_scaling)    
    return scaling_factors, star_positions









##################################################
# Diagnostic
##################################################

if __name__ == "__main__":
    fitsfile = sys.argv[1]
    foldername = fitsfile[-8:-5]
    cube = fits.getdata(fitsfile)
    nchan = cube.shape[0]
    img_shape = cube.shape[1:]
    centers = 125*np.ones((nchan, 2))
    channels = np.arange(nchan)

    #################################
    # Initial pass
    spot_masks = make_mask_grid_spots(img_shape, centers)
    spot_locs = np.zeros((num_spots, nchan, 2))
    masked_cubes =  ma.masked_array([ma.masked_array(cube, mask=i) 
                                     for i in spot_masks])
    # get reference channel grid spots:
    for i in range(num_spots):
        img = masked_cubes[i,P1640params.refchan] 
        g, spot_locs[i,P1640params.refchan] = fit_grid_spot(img, centers[P1640params.refchan]) 
    # get starting guesses for the rest
    centered_spots = spot_locs[:,P1640params.refchan] - centers[P1640params.refchan]
    init_rad = np.linalg.norm(centered_spots, axis=1)
    init_theta = np.arctan2(centered_spots[:,0], centered_spots[:,1])
    init_spots = np.rollaxis(np.rollaxis(
        np.array([np.outer(init_rad,scale_factors).T*np.sin(init_theta),
                  np.outer(init_rad,scale_factors).T*np.cos(init_theta)]), 
        2, 0), 2, 1) + centers
    
    # now, fit the rest of the spots
    spot_fits = [[] for i in range(num_spots)]
    for i in range(num_spots):
        images = masked_cubes[i] #cube*spot_masks[i]
        for chan in range(nchan):
            g, spot_locs[i,chan] = fit_grid_spot(images[chan], centers[chan], 
                                                 loc=init_spots[i,chan])
            spot_fits[i].append(g)

    #################################
    # Calculate centers
    # At each channel, fit lines through opposing spots
    for chan in range(nchan):
        b20,m20 = fit_poly(np.array((spot_locs[0,chan,1], 
                                    spot_locs[2,chan,1])),
                          np.array((spot_locs[0,chan,0], 
                                    spot_locs[2,chan,0])),
                      order=1)
        b31,m31 = fit_poly(np.array((spot_locs[1,chan,1], 
                                    spot_locs[3,chan,1])),
                          np.array((spot_locs[0,chan,0], 
                                    spot_locs[2,chan,0])),
                      order=1)
        x_center = (b31-b20)/(m20-m31)  #(line20[0]-line31[0])/(line31[1]-line20[1])
        y_center = (m31*b20-m20*b31)/(m31-m20)  #(line20[1]*line31[0]-line31[1]*line20[0])/(line20[1]-line31[1])
        centers[chan] = (y_center, x_center)

    #re-calculate positions and radii given the new centers
    centered_spots = spot_locs - centers
    rad_spots = np.linalg.norm(spot_locs-centers, axis=2)


    # remake the masks with the new centers
    #spot_masks = make_mask_grid_spots(img_shape, centers)





import warnings
import pickle
import math
import sys

import numpy as np
import scipy.linalg as linalg
import scipy.ndimage as ndi
import scipy.ndimage.interpolation as sinterp
import scipy.optimize as optimize

import pyklip.covars as covars
import astropy.stats.circstats as circstats

# emcee more MCMC sampling
import emcee

#Check python version
if sys.version_info < (3,0):
    v2 = True
else:
    v2 = False
    
#import pymultinest if the user has it installed
if v2 == False:
    try:
        import pymultinest
        nomultinest = False
    except ModuleNotFoundError:
        nomultinest = True
elif v2 == True:
    try:
        import pymultinest
        nomultinest = False
    except ImportError:
        nomultinest = True

class FitPSF(object):
    """
    Base class to perform astrometry on direct imaging data_stamp using GP regression. Can utilize a Bayesian framework with MCMC or a frequentist framework with least squares. 

    Args:
        fitboxsize: fitting box side length (pixels)
        method (str): either 'mcmc' or 'maxl' depending on framework you want. Defaults to 'mcmc'. 
        fmt (str): either 'seppa' or 'xy' depending on how you want to input the guess coordiantes

    Attributes:
        guess_x (float): (initialization) guess x position [pixels]
        guess_y (float): (initialization) guess y positon [pixels]
        guess_flux (float): guess scale factor between model and data
        fit_x (:py:class:`pyklip.fitpsf.ParamRange`): (result) the result from the MCMC fit for the planet's location [pixels]
        fit_y (:py:class:`pyklip.fitpsf.ParamRange`): (result) the result from the MCMC fit for the planet's location [pixels]
        fit_flux (:py:class:`pyklip.fitpsf.ParamRange`): (result) factor to scale the FM to match the flux of the data
        covar_params (list of :py:class:`pyklip.fitpsf.ParamRange`): (result) hyperparameters for the Gaussian processa
        fm_stamp (np.array): (fitting) The 2-D stamp of the forward model (centered at the nearest pixel to the guessed location)
        data_stamp (np.array): (fitting) The stamp of the data (centered at the nearest pixel to the guessed location) (2-D unless there were NaNs in which 1-D)
        noise_map (np.array): (fitting) The stamp of the noise for each pixel the data computed assuming azimuthally similar noise (same dim as data_stamp)
        padding (int): amount of pixels on one side to pad the data/forward model stamp
        sampler (emcee.EnsembleSampler): an instance of the emcee EnsambleSampler. Only for Bayesian fit. See emcee docs for more details. 
    

    """
    def __init__(self, fitboxsize, method='mcmc'):
        """
        Initilaizes the FitPSF class
        """
        # store initailization
        self.fitboxsize = fitboxsize
        if method.lower() == "maxl":
            self.isbayesian = False
        elif method.lower() == "mcmc":
            self.isbayesian = True
        else:
            raise ValueError("method needs to be either 'maxl' or 'mcmc'. Received {0}.".format(method))


        # stuff that isn't generated yet
        # stamps of the data_stamp and the forward model
        self.fm_stamp = None # Forward Model
        self.padding = 0 # padding for FM. You kinda need this to shift the FM around
        self.data_stamp = None # Data
        self.noise_map = None # same shape as self.data_stamp
        self.data_stamp_x = None # RA offset of data_stamp (in pixels)
        self.data_stamp_y = None # Dec offset (in pixels)
        self.data_stamp_x_center = None # RA offset of center pixel (stampsize // 2)
        self.data_stamp_y_center = None # Dec offset of center pixel (stampsize // 2)

        # guess flux (a hyperparameter)
        self._guess_flux = None

        # covariance paramters. Use the covariance initilizer function to initilize them
        self.covar = None
        self.covar_param_guesses = None
        self.covar_param_labels = None
        self.include_readnoise = False

        # MCMC fit params
        self.bounds = None
        self.sampler = None
        # Max-Likelihood fit
        self.hess_inv = None

        # best fit
        self.fit_x = None
        self.fit_y = None
        self.fit_flux = None
        self.covar_params = None

    # automatically guess flux if it hasn't been defined
    @property
    def guess_flux(self):
        # if it was already set return it
        if self._guess_flux is not None:
            return self._guess_flux
        # if the data and fm stamps haven't been created yet, we can't do much 
        elif self.data_stamp is None or self.fm_stamp is None:
            return None
        # guess it based on the max value in both stamps
        self._guess_flux = np.nanmax(self.data_stamp) / np.nanmax(self.fm_stamp)
        return self._guess_flux
    @guess_flux.setter
    def guess_flux(self, newval):
        self._guess_flux = newval

    def generate_fm_stamp(self, fm_image, fm_pos=None, fm_wcs=None, extract=True, padding=5):
        """
        Generates a stamp of the forward model and stores it in self.fm_stamp
        Args:
            fm_image: full image containing the fm_stamp
            fm_pos: [x,y] location of the forwrd model in the fm_image
            fm_wcs: if not None, specifies the sky angles in the image. If None, assume image is North up East left
            extract: if True, need to extract the forward model from the image. Otherwise, assume the fm_stamp is already
                    centered in the frame (fm_image.shape // 2)
            padding: number of pixels on each side in addition to the fitboxsize to extract to pad the fm_stamp
                        (should be >= 1)

        Returns:

        """
        # cheeck the padding to make sure it's valid
        if not isinstance(padding, int):
            raise TypeError("padding must be an integer")
        if padding < 1:
            warnings.warn("Padding really should be >= 1 pixel so we can shift the FM around", RuntimeWarning)
        self.padding = padding


        if extract:
            if fm_wcs is not None:
                raise NotImplementedError("Have not implemented rotation using WCS")

            # image is now rotated North up east left
            # find the location of the FM
            psf_xpos = fm_pos[0]
            psf_ypos = fm_pos[1]
        else:
            # PSf is already cenetered
            psf_xpos = fm_image.shape[1]//2
            psf_ypos = fm_image.shape[0]//2

        # now we found the FM in the image, extract out a centered stamp of it
        # grab the coordinates of the image
        stampsize = 2 * self.padding + self.fitboxsize # full stamp needs padding around all sides
        x_stamp, y_stamp = np.meshgrid(np.arange(stampsize * 1.) - stampsize //2,
                                       np.arange(stampsize * 1.) - stampsize// 2)

        x_stamp += psf_xpos
        y_stamp += psf_ypos

        # zero nans because it messes with interpolation
        fm_image[np.where(np.isnan(fm_image))] = 0

        fm_stamp = ndi.map_coordinates(fm_image, [y_stamp, x_stamp])
        self.fm_stamp = fm_stamp



    def generate_data_stamp(self, data, guess_loc, noise_map, radial_noise_center=None, dr=4, exclusion_radius=10):
        """
        Generate a stamp of the data_stamp ~centered on planet and also corresponding noise map
        Args:
            data: the final collapsed data_stamp (2-D)
            guess_loc: guess location of where to fit the model in the data 
            noise_map: if not None, noise map for each pixel (either same shape as input data, or shape of data stamp)
                       if None, one will be generated assuming azimuthal noise using an annulus widthh of dr. radial_noise_center MUST be defined.
            radial_noise_center: if we assume the noise is azimuthally symmetric and changes radially, this is the [x,y] center for it
            dr: width of annulus in pixels from which the noise map will be generated
            exclusion_radius: radius around the guess planet location which doens't get factored into the radial noise estimate

        Returns:

        """
        if noise_map is None and radial_noise_center is None:
            raise ValueError("radial_noise_center needs to be specified if noise map is not passed in")
        
        # store initailization
        self.guess_x = guess_loc[0]
        self.guess_y = guess_loc[1]

        # round to nearest pixel
        xguess_round = int(np.round(self.guess_x))
        yguess_round = int(np.round(self.guess_y))

        # get index bounds for grabbing pixels from data_stamp
        ymin = yguess_round - self.fitboxsize//2
        xmin = xguess_round - self.fitboxsize//2
        ymax = yguess_round + self.fitboxsize//2 + 1
        xmax = xguess_round + self.fitboxsize//2 + 1
        if self.fitboxsize % 2 == 0:
            # for even fitbox sizes, need to truncate ymax/xmax by 1
            ymax -= 1
            xmax -= 1

        data_stamp = data[ymin:ymax, xmin:xmax]
        self.data_stamp = data_stamp

        # store coordinates of stamp also
        y_img, x_img = np.indices(data.shape, dtype=float)

        x_data_stamp = x_img[ymin:ymax, xmin:xmax]
        y_data_stamp = y_img[ymin:ymax, xmin:xmax]
        self.data_stamp_x = x_data_stamp
        self.data_stamp_y = y_data_stamp
        self.data_stamp_x_center = self.data_stamp_x[0, self.fitboxsize // 2]
        self.data_stamp_y_center = self.data_stamp_y[self.fitboxsize // 2, 0]

        if noise_map is not None:
            # check size of noise map:
            if noise_map.shape[0] == self.fitboxsize and noise_map.shape[1] == self.fitboxsize:
                noise_stamp = noise_map
            else:
                # assume it is the whole image in size
                noise_stamp = noise_map[ymin:ymax, xmin:xmax]
        else:
            # need to generate the noise map assuming noise is azimuthally symmetric about a center
            # blank map
            noise_stamp = np.zeros(data_stamp.shape)

            # define exclusion around planet.
            distance_from_planet = np.sqrt((x_img - self.guess_x)**2 + (y_img - self.guess_y)**2)
            # define radial coordinate
            r_img = np.sqrt((x_img - radial_noise_center[0])**2 + (y_img - radial_noise_center[1])**2)
            r_stamp = np.sqrt((x_data_stamp - radial_noise_center[0])**2 + (y_data_stamp - radial_noise_center[1])**2)

            # calculate noise for each pixel in the data_stamp stamp
            for y_index, x_index in np.ndindex(data_stamp.shape):
                r_pix = r_stamp[y_index, x_index]
                pixels_for_noise = np.where((np.abs(r_img - r_pix) <= dr/2.) & (distance_from_planet > exclusion_radius))
                noise_stamp[y_index, x_index] = np.nanstd(data[pixels_for_noise])

        self.noise_map = noise_stamp

        # if there are NaNs, unravel the data to 1-D and remove NaNs
        nanpix = np.where(np.isnan(self.data_stamp))
        if np.size(nanpix) > 0:
            goodpix = np.where(~np.isnan(self.data_stamp))
            # self.data_stamp = self.data_stamp[goodpix]
            # self.data_stamp_RA_offset = self.data_stamp_RA_offset[goodpix]
            # self.data_stamp_Dec_offset = self.data_stamp_Dec_offset[goodpix]
            # self.noise_map = self.noise_map[goodpix]
            self._usegoodpix = goodpix
        else:
            self._usegoodpix = None


    def set_kernel(self, covar, covar_param_guesses, covar_param_labels, include_readnoise=False,
                   read_noise_fraction=0.01):
        """
        Set the Gaussian process kernel used in our fit

        Args:
            covar: Covariance kernel for GP regression. If string, can be "matern32" or "sqexp" or "diag"
                    Can also be a function: cov = cov_function(x_indices, y_indices, sigmas, cov_params)
            covar_param_guesses: a list of guesses on the hyperparmeteres (size of N_hyperparams). This can be an empty list for 'diag'.
            covar_param_labels: a list of strings labelling each covariance parameter
            include_readnoise: if True, part of the noise is a purely diagonal term (i.e. read/photon noise)
            read_noise_fraction: fraction of the total measured noise is read noise (between 0 and 1)

        Returns:

        """
        if isinstance(covar, str):
            if covar.lower() == "matern32":
                self.covar = covars.matern32
            elif covar.lower() == "sqexp":
                self.covar = covars.sq_exp
            elif covar.lower() == "diag":
                self.covar = covars.delta
            else:
                raise ValueError("Covariance matricies currently supported are 'matern32', 'sqexp', and diag")
        else:
            # this better be a covariance function. We're trusting you
            self.covar = covar

        self.covar_param_guesses = covar_param_guesses
        self.covar_param_labels = covar_param_labels

        if include_readnoise:
            self.include_readnoise = True
            self.covar_param_guesses.append(read_noise_fraction)
            self.covar_param_labels.append(r"K_{\delta}")


    def set_bounds(self, dx, dy, df, covar_param_bounds, read_noise_bounds=None):
        """
        Set bounds on Bayesian priors. All paramters can be a 2 element tuple/list/array that specifies
        the lower and upper bounds x_min < x < x_max. Or a single value whose interpretation is specified below
        If you are passing in both lower and upper bounds, both should be in linear scale!
        Args:
            dx: Distance from initial guess position in pixels. For a single value, this specifies the largest distance
                form the initial guess (i.e. x_guess - dx < x < x_guess + dx)
            dy: Same as dx except with y
            df: Flux range. If single value, specifies how many orders of 10 the flux factor can span in one direction
                (i.e. log_10(guess_flux) - df < log_10(guess_flux) < log_10(guess_flux) + df
            covar_param_bounds: Params for covariance matrix. Like df, single value specifies how many orders of
                                magnitude parameter can span. Otherwise, should be a list of 2-elem touples
            read_noise_bounds: Param for read noise term. If single value, specifies how close to 0 it can go
                                based on powers of 10 (i.e. log_10(-read_noise_bound) < read_noise < 1 )

        Returns:

        """
        self.bounds = []

        # x/RA bounds
        if np.size(dx) == 2:
            self.bounds.append(dx)
        else:
            self.bounds.append([self.guess_x - dx, self.guess_x + dx])

        # y/Dec bounds
        if np.size(dy) == 2:
            self.bounds.append(dy)
        else:
            self.bounds.append([self.guess_y - dy, self.guess_y + dy])

        if np.size(df) == 2:
            self.bounds.append(df)
        else:
            self.bounds.append([self.guess_flux / (10.**df), self.guess_flux * (10**df)])

        # hyperparam bounds
        if np.ndim(covar_param_bounds) == 2:
            for covar_param_bound in covar_param_bounds:
                self.bounds.append(covar_param_bound)
        else:
            # this is a 1-D list, with each param specified by one paramter
            for covar_param_bound, covar_param_guess in zip(covar_param_bounds, self.covar_param_guesses):
                self.bounds.append([covar_param_guess / (10.**covar_param_bound),
                                    covar_param_guess * (10**covar_param_bound)])

        if read_noise_bounds is not None:
        # read noise
            if np.size(read_noise_bounds) == 2:
                self.bounds.append(read_noise_bounds)
            else:
                self.bounds.append([self.covar_param_guesses[-1]/10**read_noise_bounds, 1])


    def fit_psf(self, nwalkers=100, nburn=200, nsteps=800, save_chain=True, chain_output="bka-chain.pkl",
                       numthreads=None):
        """
        Fits the PSF to the data in either a frequentist or Bayesian way depending on initialization. 

        Args:
            nwalkers: number of walkers (mcmc-only)
            nburn: numbe of samples of burn-in for each walker (mcmc-only)
            nsteps: number of samples each walker takes (mcmc-only)
            save_chain: if True, save the output in a pickled file (mcmc-only)
            chain_output: filename to output the chain to (mcmc-only)
            numthreads: number of threads to use (mcmc-only)

        Returns:

        """
        if self.isbayesian:
            return self._mcmc_fit_psf(nwalkers=nwalkers, nburn=nburn, nsteps=nsteps, save_chain=save_chain, 
                                        chain_output=chain_output, numthreads=numthreads)
        else:
            return self._lsqr_fit_psf()   

    def _mcmc_fit_psf(self, nwalkers=100, nburn=200, nsteps=800, save_chain=True, chain_output="bka-chain.pkl",
                       numthreads=None):
        """
        Run a Bayesian fit of the astrometry using MCMC
        Saves to self.chian

        Args:
            nwalkers: number of walkers
            nburn: numbe of samples of burn-in for each walker
            nsteps: number of samples each walker takes
            save_chain: if True, save the output in a pickled file
            chain_output: filename to output the chain to
            numthreads: number of threads to use

        Returns:

        """
        # create array of initial guesses
        # array of guess RA, Dec, and flux
        # for everything that's not RA/Dec offset, should be converted to log space for MCMC sampling
        init_guess = np.array([self.guess_x, self.guess_y, math.log(self.guess_flux)])
        # append hyperparams for covariance matrix, which also need to be converted to log space
        init_guess = np.append(init_guess, np.log(self.covar_param_guesses))
        # number of dimensions of MCMC fit
        ndim = np.size(init_guess)

        # initialize walkers in a ball around the best fit value
        pos = [init_guess + 1e-4 * np.random.randn(ndim) for i in range(nwalkers)]

        # prior bounds also need to be put in log space
        sampler_bounds = np.copy(self.bounds)
        sampler_bounds[2:] = np.log(sampler_bounds[2:])

        global lnprob
        sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob, args=(self, sampler_bounds, self.covar),
                                        kwargs={'readnoise' : self.include_readnoise}, threads=numthreads)

        # burn inf
        print("Running burn in")
        pos, _, _ = sampler.run_mcmc(pos, nburn)
        # reset sampler
        sampler.reset()
         
        # chains should hopefulyl have converged. Now run MCMC
        print("Burn in finished. Now sampling posterior")
        sampler.run_mcmc(pos, nsteps)
        print("MCMC sampler has finished")

        # convert chains in log space back in linear space
        sampler.chain[:,:,2:] = np.exp(sampler.chain[:,:,2:])

        # save state
        self.sampler = sampler

        # save best fit values
        # percentiles has shape [ndims, 3]
        percentiles = np.swapaxes(np.percentile(sampler.flatchain, [16, 50, 84], axis=0), 0, 1)
        self.fit_x = ParamRange(percentiles[0][1], np.array([percentiles[0][2], percentiles[0][0]]) - percentiles[0][1])
        self.fit_y = ParamRange(percentiles[1][1], np.array([percentiles[1][2], percentiles[1][0]]) - percentiles[1][1])
        self.fit_flux =  ParamRange(percentiles[2][1], np.array([percentiles[2][2], percentiles[2][0]]) -  percentiles[2][1])
        self.covar_params = [ParamRange(thispercentile[1], np.array([thispercentile[2], thispercentile[0]]) - thispercentile[1] ) for thispercentile in percentiles[3:]]

        if save_chain:
            pickle_file = open(chain_output, 'wb')
            pickle.dump(sampler.chain, pickle_file)
            pickle.dump(sampler.lnprobability, pickle_file)
            pickle.dump(sampler.acceptance_fraction, pickle_file)
            #pickle.dump(sampler.acor, pickle_file)
            pickle_file.close()

    def _lsqr_fit_psf(self):
        """
        Do a frequentist maximum likelihood fit to the data. Approximate errors using the Hessian of the likelihood function.
        """
        # create array of initial guesses
        # array of guess RA, Dec, and flux
        # for everything that's not RA/Dec offset, should be converted to log space for MCMC sampling
        # init_guess = np.array([self.guess_x, self.guess_y, math.log(self.guess_flux)])
        # # append hyperparams for covariance matrix, which also need to be converted to log space
        # init_guess = np.append(init_guess, np.log(self.covar_param_guesses))

        init_guess = np.array([self.guess_x, self.guess_y, (self.guess_flux)])
        # append hyperparams for covariance matrix, which also need to be converted to log space
        init_guess = np.append(init_guess, (self.covar_param_guesses))

        if self.bounds is None:
            cost_function = lnprob
            # construct some bounds, just very loose
            bounds = [[self.guess_x - self.fitboxsize/2, self.guess_x + self.fitboxsize/2], [self.guess_y - self.fitboxsize/2, self.guess_y + self.fitboxsize/2],
                      [0, np.inf]]
            for _ in self.covar_param_guesses:
                bounds += [[0, np.inf]]
            cost_function_args = (self, bounds, self.covar)
            self.bounds = bounds
        else:
            # prior bounds also need to be put in log space
            sampler_bounds = np.copy(self.bounds)
            #sampler_bounds[2:] = np.log(sampler_bounds[2:])

            cost_function = lnprob
            cost_function_args = (self, sampler_bounds, self.covar)

        cost_function_args += (self.include_readnoise, True)


        #global cost_function
        nm_result = optimize.minimize(cost_function, init_guess, args=cost_function_args, method="Nelder-Mead")

        # BFGS will only fit for position and flux, and their uncertainties.
        new_init_guess = nm_result.x[:3]
        new_init_guess = np.append(new_init_guess, nm_result.x[3:])
        #if cost_function_args[1] is not None:
        #    cost_function_args[1] = cost_function_args[1][:3] # modify limits to not include hyperparameters 

        result = optimize.minimize(cost_function, new_init_guess, args=cost_function_args, method="BFGS")

        if not result.success:
            warnings.warn("Optimizer did not converge! Estimated uncertainties are likely unreliable. Msg: {0}".format(result.message))
            
        # best fit values, and use the Hessian to approximate the uncertainties in the parameters
        ra_best = result.x[0]
        dec_best = result.x[1]
        flux_best = (result.x[2])
        covar_params_best = [nm_result.x[i] for i in range(3, np.size(nm_result.x))]
        ra_err = np.sqrt(np.abs(result.hess_inv[0,0]))
        dec_err = np.sqrt(np.abs(result.hess_inv[1,1]))
        flux_err = np.sqrt(np.abs(result.hess_inv[2,2]))

        # convert to linear space the flux and covariance parameters
        covar_params_best = np.exp(covar_params_best)
        # flux error is approx symmetric in log space
        flux_err_two_sided = np.array([np.exp(flux_best - flux_err), np.exp(flux_best + flux_err)])
        flux_best = math.exp(flux_best)

        # save best fit values
        # percentiles has shape [ndims, 3]
        self.fit_x = ParamRange(ra_best, ra_err)
        self.fit_y = ParamRange(dec_best, dec_err)
        self.fit_flux =  ParamRange(flux_best, flux_err_two_sided)
        self.covar_params = [ParamRange(param_best, 0) for param_best in covar_params_best]

        self.hess_inv = result.hess_inv
        

    def make_corner_plot(self, fig=None):
        """
        Generate a corner plot of the posteriors from the MCMC
        Args:
            fig: if not None, a matplotlib Figure object

        Returns:
            fig: the Figure object. If input fig is None, function will make a new one

        """
        if not self.isbayesian:
            raise AttributeError("Corner plot is only available if using Bayesian MCMC framework")
        
        import corner

        all_labels = [r"x", r"y", r"$\alpha$"]
        all_labels = np.append(all_labels, self.covar_param_labels)

        fig = corner.corner(self.sampler.flatchain, labels=all_labels, quantiles=[0.16, 0.5, 0.84], fig=fig)

        return fig


    def best_fit_and_residuals(self, fig=None):
        """
        Generate a plot of the best fit FM compared with the data_stamp and also the residuals
        Args:
            fig (matplotlib.Figure): if not None, a matplotlib Figure object

        Returns:
            fig (matplotlib.Figure): the Figure object. If input fig is None, function will make a new one

        """
        import matplotlib
        import matplotlib.pylab as plt

        if fig is None:
            fig = plt.figure(figsize=(12, 4))

        # create best fit FM
        dx = self.fit_x.bestfit - self.data_stamp_x_center
        dy = self.fit_y.bestfit - self.data_stamp_y_center

        fm_bestfit = self.fit_flux.bestfit * sinterp.shift(self.fm_stamp, [dy, dx])
        if self.padding > 0:
            fm_bestfit = fm_bestfit[self.padding:-self.padding, self.padding:-self.padding]

        # make residual map
        residual_map = self.data_stamp - fm_bestfit

        # normalize all images to same scale
        colornorm = matplotlib.colors.Normalize(vmin=np.nanpercentile(self.data_stamp, 0.03),
                                                vmax=np.nanpercentile(self.data_stamp, 99.7))

        # plot the data_stamp
        ax1 = fig.add_subplot(131)
        im1 = ax1.imshow(self.data_stamp, interpolation='nearest', cmap='cubehelix', norm=colornorm)
        ax1.invert_yaxis()
        ax1.set_title("Data")
        ax1.set_xlabel("X (pixels)")
        ax1.set_ylabel("Y (pixels)")

        ax2 = fig.add_subplot(132)
        im2 = ax2.imshow(fm_bestfit, interpolation='nearest', cmap='cubehelix', norm=colornorm)
        ax2.invert_yaxis()
        ax2.set_title("Best-fit Model")
        ax2.set_xlabel("X (pixels)")

        ax3 = fig.add_subplot(133)
        im3 = ax3.imshow(residual_map, interpolation='nearest', cmap='cubehelix', norm=colornorm)
        ax3.invert_yaxis()
        ax3.set_title("Residuals")
        ax3.set_xlabel("X (pixels)")

        fig.subplots_adjust(right=0.82)
        fig.subplots_adjust(hspace=0.4)
        ax_pos = ax3.get_position()

        cbar_ax = fig.add_axes([0.84, ax_pos.y0, 0.02, ax_pos.height])
        cb = fig.colorbar(im1, cax=cbar_ax)
        cb.set_label("Counts (DN)")

        return fig


def lnprior(fitparams, bounds, readnoise=False, negate=False):
    """
    Bayesian prior

    Args:
        fitparams: array of params (size N)

        bounds: array of (N,2) with corresponding lower and upper bound of params
                bounds[i,0] <= fitparams[i] < bounds[i,1]
        readnoise (bool): If True, the last fitparam fits for diagonal noise
        negate (bool): if True, negatives the probability (used for minimization algos)

    Returns:
        prior: 0 if inside bound ranges, -inf if outside

    """
    prior = 0.0

    for param, bound in zip(fitparams, bounds):
        if (param >= bound[1]) | (param < bound[0]):
            prior *= -np.inf
            break

    if negate:
        prior *= -1

    return prior


def lnlike(fitparams, fma, cov_func, readnoise=False, negate=False):
    """
    Likelihood function
    Args:
        fitparams: array of params (size N). First three are [dRA,dDec,f]. Additional parameters are GP hyperparams
                    dRA,dDec: RA,Dec offsets from star. Also coordianaes in self.data_{RA,Dec}_offset
                    f: flux scale factor to normalizae the flux of the data_stamp to the model
        fma (FMAstrometry): a FMAstrometry object that has been fully set up to run
        cov_func (function): function that given an input [x,y] coordinate array returns the covariance matrix
                  e.g. cov = cov_function(x_indices, y_indices, sigmas, cov_params)
        readnoise (bool): If True, the last fitparam fits for diagonal noise
        negate (bool): if True, negatives the probability (used for minimization algos)

    Returns:
        likeli: log of likelihood function (minus a constant factor)
    """
    x_trial = fitparams[0]
    y_trial = fitparams[1]
    f_trial = fitparams[2]
    hyperparms_trial = fitparams[3:]

    if readnoise:
        # last hyperparameter is a diagonal noise term. Separate it out
        readnoise_amp = np.exp(hyperparms_trial[-1])
        hyperparms_trial = hyperparms_trial[:-1]

    # get trial parameters out of log space
    f_trial = math.exp(f_trial)
    hyperparms_trial = np.exp(hyperparms_trial)

    dx = x_trial - fma.data_stamp_x_center
    dy = y_trial - fma.data_stamp_y_center

    fm_shifted = sinterp.shift(fma.fm_stamp, [dy, dx])

    if fma.padding > 0:
        fm_shifted = fm_shifted[fma.padding:-fma.padding, fma.padding:-fma.padding]

    if fma._usegoodpix is not None:
        fm_shifted = fm_shifted[fma._usegoodpix]
        data_stamp = fma.data_stamp[fma._usegoodpix]
        x_offsets = fma.data_stamp_x[fma._usegoodpix]
        y_offsets = fma.data_stamp_y[fma._usegoodpix]
        noise_map = fma.noise_map[fma._usegoodpix]
    else:
        data_stamp = fma.data_stamp
        x_offsets = fma.data_stamp_x
        y_offsets = fma.data_stamp_y
        noise_map = fma.noise_map

    diff_ravel = data_stamp.ravel() - f_trial * fm_shifted.ravel()

    cov = cov_func(x_offsets.ravel(), y_offsets.ravel(), noise_map.ravel(), hyperparms_trial)

    if readnoise:
        # add a diagonal term
        cov = (1 - readnoise_amp) * cov + readnoise_amp * np.diagflat(noise_map.ravel()**2 )
    
    # solve Cov * x = diff for x = Cov^-1 diff. Numerically more stable than inverse
    # to make it faster, we comptue the Cholesky factor and use it to also compute the determinent
    try:
        (L_cov, lower_cov) = linalg.cho_factor(cov)
        cov_inv_dot_diff = linalg.cho_solve((L_cov, lower_cov), diff_ravel) # solve Cov x = diff for x
        logdet = 2*np.sum(np.log(np.diag(L_cov)))
    except: 
        cov_inv = np.linalg.inv(cov)
        cov_inv_dot_diff = np.dot(cov_inv, diff_ravel)
        logdet = np.linalg.slogdet(cov)[1]
    
    residuals = diff_ravel.dot(cov_inv_dot_diff)
    constant = logdet

    loglikelihood = -0.5 * (residuals + constant)

    if negate:
        loglikelihood *= -1
    
    return loglikelihood


def lnprob(fitparams, fma, bounds, cov_func, readnoise=False, negate=False):
    """
    Function to compute the relative posterior probabiltiy. Product of likelihood and prior
    Args:
        fitparams: array of params (size N). First three are [dRA,dDec,f]. Additional parameters are GP hyperparams
                    dRA,dDec: RA,Dec offsets from star. Also coordianaes in self.data_{RA,Dec}_offset
                    f: flux scale factor to normalizae the flux of the data_stamp to the model
        fma: a FMAstrometry object that has been fully set up to run
        bounds: array of (N,2) with corresponding lower and upper bound of params
                bounds[i,0] <= fitparams[i] < bounds[i,1]
        cov_func: function that given an input [x,y] coordinate array returns the covariance matrix
                  e.g. cov = cov_function(x_indices, y_indices, sigmas, cov_params)
        readnoise (bool): If True, the last fitparam fits for diagonal noise
        negate (bool): if True, negatives the probability (used for minimization algos)

    Returns:

    """
    lp = lnprior(fitparams, bounds, readnoise=readnoise, negate=negate)

    if not np.isfinite(lp):
        if not negate:
            return -np.inf
        else:
            return np.inf
    return lp + lnlike(fitparams, fma, cov_func, readnoise=readnoise, negate=negate)


class ParamRange(object):
    """
    Stores the best fit value and uncertainities for a parameter in a neat fasion

    Args:
        bestfit (float): the bestfit value
        err_range: either a float or a 2-element tuple (+val1, -val2) and gives the 1-sigma range

    Attributes:
        bestfit (float): the bestfit value
        error (float): the average 1-sigma error
        error_2sided (np.array): [+error1, -error2] 2-element array with asymmetric errors
    """
    def __init__(self, bestfit, err_range):
        self.bestfit = bestfit

        if isinstance(err_range, (int, float)):
            self.error = err_range
            self.error_2sided = np.array([err_range, -err_range])
        elif len(err_range) == 2:
            self.error_2sided = np.array(err_range)
            self.error = np.mean(np.abs(err_range))

class FMAstrometry(FitPSF):
    """
    Specifically for fitting astrometry of a directly imaged companion relative to its star. Extension of :py:class:`pyklip.fitpsf.FitPSF`.

    Args:
        guess_sep: the guessed separation (pixels)
        guess_pa: the guessed position angle (degrees)
        fitboxsize: fitting box side length (pixels)
        method (str): either 'mcmc' or 'maxl' depending on framework you want. Defaults to 'mcmc'. 

    Attributes:
        guess_sep (float): (initialization) guess separation for planet [pixels]
        guess_pa (float): (initialization) guess PA for planet [degrees]
        guess_RA_offset (float): (initialization) guess RA offset [pixels]
        guess_Dec_offset (float): (initialization) guess Dec offset [pixels]
        raw_RA_offset (:py:class:`pyklip.fitpsf.ParamRange`): (result) the raw result from the MCMC fit for the planet's location [pixels]
        raw_Dec_offset (:py:class:`pyklip.fitpsf.ParamRange`): (result) the raw result from the MCMC fit for the planet's location [pixels]
        raw_flux (:py:class:`pyklip.fitpsf.ParamRange`): (result) factor to scale the FM to match the flux of the data
        covar_params (list of :py:class:`pyklip.fitpsf.ParamRange`): (result) hyperparameters for the Gaussian process
        raw_sep(:py:class:`pyklip.fitpsf.ParamRange`): (result) the inferred raw result from the MCMC fit for the planet's location [pixels]
        raw_PA(:py:class:`pyklip.fitpsf.ParamRange`): (result) the inferred raw result from the MCMC fit for the planet's location [degrees]
        RA_offset(:py:class:`pyklip.fitpsf.ParamRange`): (result) the RA offset of the planet that includes all astrometric errors [pixels or mas]
        Dec_offset(:py:class:`pyklip.fitpsf.ParamRange`): (result) the Dec offset of the planet that includes all astrometric errors [pixels or mas]
        sep(:py:class:`pyklip.fitpsf.ParamRange`): (result) the separation of the planet that includes all astrometric errors [pixels or mas]
        PA(:py:class:`pyklip.fitpsf.ParamRange`): (result) the PA of the planet that includes all astrometric errors [degrees]
        fm_stamp (np.array): (fitting) The 2-D stamp of the forward model (centered at the nearest pixel to the guessed location)
        data_stamp (np.array): (fitting) The 2-D stamp of the data (centered at the nearest pixel to the guessed location)
        noise_map (np.array): (fitting) The 2-D stamp of the noise for each pixel the data computed assuming azimuthally similar noise
        padding (int): amount of pixels on one side to pad the data/forward model stamp
        sampler (emcee.EnsembleSampler): an instance of the emcee EnsambleSampler. Only for Bayesian fit. See emcee docs for more details. 
    """
    def __init__(self, guess_sep, guess_pa, fitboxsize, method='mcmc'):

        # derive delta RA and delta Dec
        # in pixels
        self.guess_sep = guess_sep
        self.guess_pa = guess_pa

        self.guess_RA_offset = self.guess_sep * np.sin(np.radians(self.guess_pa))
        self.guess_Dec_offset = self.guess_sep * np.cos(np.radians(self.guess_pa))

        # best fit
        self.raw_RA_offset = None
        self.raw_Dec_offset = None
        self.raw_flux = None
        # best fit infered parameters
        self.raw_sep = None
        self.raw_PA = None
        self.RA_offset = None
        self.Dec_offset = None
        self.sep = None
        self.PA = None

        super(FMAstrometry, self).__init__(fitboxsize, method)

    def generate_fm_stamp(self, fm_image, fm_center, fm_wcs=None, extract=True, padding=5):
        """
        Generates a stamp of the forward model and stores it in self.fm_stamp
        Args:
            fm_image: full imgae containing the fm_stamp
            fm_center: [x,y] center of image (assuing fm_stamp is located at sep/PA) corresponding to guess_sep and guess_pa
            fm_wcs: if not None, specifies the sky angles in the image. If None, assume image is North up East left
            extract: if True, need to extract the forward model from the image. Otherwise, assume the fm_stamp is already
                    centered in the frame (fm_image.shape // 2)
            padding: number of pixels on each side in addition to the fitboxsize to extract to pad the fm_stamp
                        (should be >= 1)

        Returns:

        """
        fm_x = -self.guess_RA_offset + fm_center[0]
        fm_y = self.guess_Dec_offset + fm_center[1]
        self.fm_center = fm_center

        super(FMAstrometry, self).generate_fm_stamp(fm_image, fm_pos=[fm_x, fm_y], fm_wcs=fm_wcs, extract=extract, padding=padding)

    def generate_data_stamp(self, data, data_center, data_wcs=None, noise_map=None, dr=4, exclusion_radius=10):
        """
        Generate a stamp of the data_stamp ~centered on planet and also corresponding noise map

        Args:
            data: the final collapsed data_stamp (2-D)
            data_center: location of star in the data_stamp.  
            data_wcs: sky angles WCS object. To rotate the image properly [NOT YET IMPLMETNED]
                      if None, data_stamp is already rotated North up East left
            noise_map: if not None, noise map for each pixel in the data_stamp (2-D).
                        if None, one will be generated assuming azimuthal noise using an annulus widthh of dr
            dr: width of annulus in pixels from which the noise map will be generated
            exclusion_radius: radius around the guess planet location which doens't get factored into noise estimate

        Returns:

        """
        # rotate image North up east left if necessary
        if data_wcs is not None:
            # rotate
            raise NotImplementedError("Rotating based on WCS is not currently implemented yet")
        xguess = -self.guess_RA_offset + data_center[0]
        yguess = self.guess_Dec_offset + data_center[1]
        self.data_center = data_center

        super(FMAstrometry, self).generate_data_stamp(data, [xguess, yguess], None, radial_noise_center=data_center, 
                                                      dr=dr, exclusion_radius=exclusion_radius)


    def fit_astrometry(self, nwalkers=100, nburn=200, nsteps=800, save_chain=True, chain_output="bka-chain.pkl",
                       numthreads=None):
        """
        Fits the PSF of the planet in either a frequentist or Bayesian way depending on initialization. 

        Args:
            nwalkers: number of walkers (mcmc-only)
            nburn: numbe of samples of burn-in for each walker (mcmc-only)
            nsteps: number of samples each walker takes (mcmc-only)
            save_chain: if True, save the output in a pickled file (mcmc-only)
            chain_output: filename to output the chain to (mcmc-only)
            numthreads: number of threads to use (mcmc-only)

        Returns:

        """
        self.fit_psf(nwalkers, nburn, nsteps, save_chain, chain_output, numthreads)
        
        # convert chains to relative separation
        self.sampler.chain[:,:,0] -= self.data_center[0]
        self.sampler.chain[:,:,0] *= -1
        self.sampler.chain[:,:,1] -= self.data_center[1]
        # save RA/dec offsets 
        self.raw_RA_offset = ParamRange(-(self.fit_x.bestfit - self.data_center[0]), self.fit_x.error_2sided[::-1])
        self.raw_Dec_offset = ParamRange(self.fit_y.bestfit - self.data_center[1], self.fit_y.error_2sided[::-1])
        self.raw_flux = self.fit_flux
    
    def propogate_errs(self, star_center_err=None, platescale=None, platescale_err=None, pa_offset=None, pa_uncertainty=None):
        """
        Propogate astrometric error. Stores results in its own fields

        Args:
            star_center_err (float): uncertainity of the star location (pixels)
            platescale (float): mas/pix conversion to angular coordinates 
            platescale_err (float): mas/pix error on the platescale
            pa_offset (float): Offset, in the same direction as position angle, to set North up (degrees)
            pa_uncertainity (float): Error on position angle/true North calibration (Degrees)
        """
        if self.isbayesian:
            # format MCMC chains
            # ensure numpy arrays
            x_fit = self.sampler.chain[:,:,0].flatten()
            y_fit = self.sampler.chain[:,:,1].flatten()

        else:
            x_fit = self.raw_RA_offset.bestfit
            y_fit = self.raw_Dec_offset.bestfit

        # calcualte statistial errors in x and y
        x_best = self.raw_RA_offset.bestfit
        y_best = self.raw_Dec_offset.bestfit
        x_1sigma_raw = self.raw_RA_offset.error_2sided
        y_1sigma_raw = self.raw_Dec_offset.error_2sided

        print("Raw X/Y Centroid = ({0}, {1}) with statistical error of {2} pix in X and {3} pix in Y".format(x_best, y_best, x_1sigma_raw, y_1sigma_raw))

        # calculate sep and pa from x/y separation
        sep_fit = np.sqrt((x_fit)**2 + (y_fit)**2)
        # For PA compute mean using circstats package, find delta_pa between all points and the mean,
        # then compute median/precentiles
        pa_fit = (np.arctan2(y_fit, -x_fit) - (np.pi/2.0)) % (2.0*np.pi) # Radians!

        if self.isbayesian:
            # for Bayesian, convert the chains to sep/pa to get uncertainity
            pa_mean = circstats.circmean(pa_fit - np.pi) + np.pi # Circmean [-pi, pi]
            d_pa = np.arctan2(np.sin(pa_fit-pa_mean), np.cos(pa_fit-pa_mean))
            pa_median = np.median(d_pa) + pa_mean
            pa_percentile = np.nanpercentile(d_pa, [84,16])  - np.median(d_pa) # median of d_pa should be small
            pa_fit = np.degrees(pa_fit) # Convert to degrees

            # calculate sep and pa statistical errors
            sep_best = np.median(sep_fit)
            pa_best = np.degrees(pa_median)
            sep_1sigma_raw = (np.percentile(sep_fit, [84,16]) - sep_best)
            pa_1sigma_raw = np.degrees(pa_percentile)
        else:
            # since we just have point estimates, use analytical error propogration
            sep_best = sep_fit
            pa_fit = np.degrees(pa_fit)
            pa_best = pa_fit

            sep_1sigma_raw = (x_fit/sep_fit)**2 * x_1sigma_raw**2 + (y_fit/sep_fit)**2 * y_1sigma_raw**2
            sep_1sigma_raw = np.sqrt(sep_1sigma_raw)

            pa_1sigma_raw = (y_fit/sep_fit**2)**2 * x_1sigma_raw**2 + (x_fit/sep_fit**2)**2 * y_1sigma_raw**2
            pa_1sigma_raw = np.sqrt(pa_1sigma_raw)
            pa_1sigma_raw = np.degrees(pa_1sigma_raw)

        print("Raw Sep/PA Centroid = ({0}, {1}) with statistical error of {2} pix in Sep and {3} pix in PA".format(sep_best, pa_best, sep_1sigma_raw, pa_1sigma_raw))
        
        # store the raw sep and PA values
        self.raw_sep = ParamRange(sep_best, sep_1sigma_raw)
        self.raw_PA = ParamRange(pa_best, pa_1sigma_raw)

        # Now let's start propogating error terms if they are supplied. 
        # We do them in Sep/PA space first since it's more natural here

        # star center error
        if star_center_err is None:
            print("Skipping star center uncertainity...")
            star_center_err = 0
        else:
            print("Adding in star center uncertainity")

        sep_err_pix = (sep_1sigma_raw**2) + star_center_err**2
        sep_err_pix = np.sqrt(sep_err_pix)

        # plate scale error
        if platescale is not None:
            print("Converting pixels to milliarcseconds")
            if platescale_err is None:
                print("Skipping plate scale uncertainity...")
                platescale_err = 0
            else:
                print("Adding in plate scale error")
            sep_err_mas = np.sqrt((sep_err_pix * platescale)**2 + (platescale_err * sep_best)**2)

        # PA Offset
        if pa_offset is not None:
            print("Adding in a PA/North angle offset")
            # Have to repeat wrapping procedure above in case pa_offset is large
            pa_fit = np.radians((pa_fit + pa_offset) % 360) # Convert back to radians for circstats
            if self.isbayesian:
                pa_mean = circstats.circmean(pa_fit - np.pi) + np.pi # Circmean [-pi, pi]
                d_pa = np.arctan2(np.sin(pa_fit-pa_mean), np.cos(pa_fit-pa_mean))

                pa_median = np.median(d_pa) + pa_mean
                pa_best = np.degrees(pa_median)
            else:
                pa_best = np.degrees(pa_fit)
            pa_fit = np.degrees(pa_fit) # Convert back to degrees
        
        # PA Uncertainity
        if pa_uncertainty is None:
            print("Skipping PA/North uncertainity...")
            pa_uncertainty = 0
        else:
            print("Adding in PA uncertainity")

        pa_err = np.radians(pa_1sigma_raw)**2 + (star_center_err/sep_best)**2 + np.radians(pa_uncertainty)**2
        pa_err = np.sqrt(pa_err)
        pa_err_deg = np.degrees(pa_err)

        sep_err_pix_avg = np.mean(np.abs(sep_err_pix))
        pa_err_deg_avg = np.mean(np.abs(pa_err_deg))

        print("Sep = {0} +/- {1} ({2}) pix, PA = {3} +/- {4} ({5}) degrees".format(sep_best, sep_err_pix_avg, sep_err_pix, pa_best, pa_err_deg_avg, pa_err_deg))

        # Store sep/PA (excluding platescale) values
        self.sep = ParamRange(sep_best, sep_err_pix)
        self.PA = ParamRange(pa_best, pa_err_deg)

        if platescale is not None:
            sep_err_mas_avg = np.mean(np.abs(sep_err_mas))
            print("Sep = {0} +/- {1} ({2}) mas, PA = {3} +/- {4} ({5}) degrees".format(sep_best*platescale, sep_err_mas_avg, sep_err_mas, pa_best, pa_err_deg_avg, pa_err_deg))
            # overwrite sep values with values converted to milliarcseconds
            self.sep = ParamRange(sep_best*platescale, sep_err_mas)

        # convert PA errors back into x y (RA/Dec)
        ra_fit = -sep_fit * np.cos(np.radians(pa_fit+90))
        dec_fit = sep_fit * np.sin(np.radians(pa_fit+90))

        # ra/dec statistical errors. This is after the rotation from pa_offset is applied
        if self.isbayesian:
            ra_best = np.median(ra_fit)
            dec_best = np.median(dec_fit)
            ra_1sigma_raw = np.percentile(ra_fit, [84,16]) - ra_best
            dec_1sigma_raw = np.percentile(dec_fit, [84,16]) - dec_best
        else:
            ra_best = ra_fit
            dec_best = dec_fit

            # this should depend on sine and cosine of pa_offset
            pa_offset_rad = np.radians(pa_offset)
            ra_1sigma_raw = np.sqrt(math.cos(pa_offset_rad)**2 * x_1sigma_raw**2 + math.sin(pa_offset_rad)**2 * y_1sigma_raw**2)
            dec_1sigma_raw = np.sqrt(math.sin(pa_offset_rad)**2 * x_1sigma_raw**2 + math.cos(pa_offset_rad)**2 * y_1sigma_raw**2)

        ra_err_full_pix = np.sqrt((ra_1sigma_raw**2)  + (star_center_err)**2 + (dec_best * np.radians(pa_uncertainty))**2 )
        dec_err_full_pix = np.sqrt((dec_1sigma_raw**2)  + (star_center_err)**2 + (ra_best * np.radians(pa_uncertainty))**2 )

        # Store error propgoated RA/Dec values (excluding platescale)
        self.RA_offset = ParamRange(ra_best, ra_err_full_pix)
        self.Dec_offset = ParamRange(dec_best, dec_err_full_pix)

        print("RA offset = {0} +/- {1} ({2}) pix".format(self.RA_offset.bestfit, self.RA_offset.error, self.RA_offset.error_2sided))
        print("Dec offset = {0} +/- {1} ({2}) pix".format(self.Dec_offset.bestfit, self.Dec_offset.error, self.Dec_offset.error_2sided))

        if platescale is not None:
            ra_err_full_mas = np.sqrt((ra_err_full_pix*platescale)**2 + (platescale_err * ra_best)**2)
            dec_err_full_mas = np.sqrt((dec_err_full_pix*platescale)**2 + (platescale_err * dec_best)**2)
            
            # Overwrite with calibrated RA/Dec converted to milliarcsecs
            self.RA_offset = ParamRange(ra_best*platescale, ra_err_full_mas)
            self.Dec_offset = ParamRange(dec_best*platescale, dec_err_full_mas)

            print("RA offset = {0} +/- {1} ({2}) mas".format(self.RA_offset.bestfit, self.RA_offset.error, self.RA_offset.error_2sided))
            print("Dec offset = {0} +/- {1} ({2}) mas".format(self.Dec_offset.bestfit, self.Dec_offset.error, self.Dec_offset.error_2sided))

def quick_psf_fit(data, psf, x_guess, y_guess, fitboxsize):
    """
    A wrapper for a quick maximum likelihood fit to a PSF to the data. 

    Args:
        data (np.array): 2-D data frame
        psf (np.array): 2-D PSF template. This should be smaller than the size of data and
                        larger than the fitboxsize
        x_guess (float): approximate x position of the location you are fitting the psf to
        y_guess (float): approximate y position of the location you are fitting the psf to
        fitboxsize (int): fitting region is a square. This is the lenght of one side of the square

    Returns:
        x_fit, y_fit, flux_fit
        x_fit (float): x position
        y_fit (float): y position
        flux_fit (float): multiplicative scale factor for the psf to match the data
    """
    fit = FitPSF(fitboxsize, method='maxl')

    padding = int((np.min(psf.shape) - fitboxsize) // 2)
    fit.generate_fm_stamp(psf, extract=False, padding=padding)

    fit.generate_data_stamp(data, [x_guess, y_guess], np.ones(data.shape))

    fit.set_kernel("diag", [], [], False)

    fit.fit_psf()

    return fit.fit_x.bestfit, fit.fit_y.bestfit, fit.fit_flux.bestfit

class PlanetEvidence(FMAstrometry):
    """
    Specifically for nested sampling of the parameter space of a directly imaged companion relative to its star. Extension of :py:class:`pyklip.fitpsf.FitPSF`.

    Args:
        guess_sep: the guessed separation (pixels)
        guess_pa: the guessed position angle (degrees)
        fitboxsize: fitting box side length (pixels)
        fm_basename: Prefix of the foward model sampling files multinest saves in /chains/
        null_basename: Prefix of the null hypothesis model sampling files multinest saves in /chains/
    
    Attributes:
        guess_sep (float): (initialization) guess separation for planet [pixels]
        guess_pa (float): (initialization) guess PA for planet [degrees]
        guess_RA_offset (float): (initialization) guess RA offset [pixels]
        guess_Dec_offset (float): (initialization) guess Dec offset [pixels]
        raw_RA_offset (:py:class:`pyklip.fitpsf.ParamRange`): (result) the raw result from the MCMC fit for the planet's location [pixels]
        raw_Dec_offset (:py:class:`pyklip.fitpsf.ParamRange`): (result) the raw result from the MCMC fit for the planet's location [pixels]
        raw_flux (:py:class:`pyklip.fitpsf.ParamRange`): (result) factor to scale the FM to match the flux of the data
        covar_params (list of :py:class:`pyklip.fitpsf.ParamRange`): (result) hyperparameters for the Gaussian process
        raw_sep(:py:class:`pyklip.fitpsf.ParamRange`): (result) the inferred raw result from the MCMC fit for the planet's location [pixels]
        raw_PA(:py:class:`pyklip.fitpsf.ParamRange`): (result) the inferred raw result from the MCMC fit for the planet's location [degrees]
        RA_offset(:py:class:`pyklip.fitpsf.ParamRange`): (result) the RA offset of the planet that includes all astrometric errors [pixels or mas]
        Dec_offset(:py:class:`pyklip.fitpsf.ParamRange`): (result) the Dec offset of the planet that includes all astrometric errors [pixels or mas]
        sep(:py:class:`pyklip.fitpsf.ParamRange`): (result) the separation of the planet that includes all astrometric errors [pixels or mas]
        PA(:py:class:`pyklip.fitpsf.ParamRange`): (result) the PA of the planet that includes all astrometric errors [degrees]
        fm_stamp (np.array): (fitting) The 2-D stamp of the forward model (centered at the nearest pixel to the guessed location)
        data_stamp (np.array): (fitting) The 2-D stamp of the data (centered at the nearest pixel to the guessed location)
        noise_map (np.array): (fitting) The 2-D stamp of the noise for each pixel the data computed assuming azimuthally similar noise
        padding (int): amount of pixels on one side to pad the data/forward model stamp
        sampler (pymultinest.run): function that runs the pymultinest sampling for both hypotheses
    """
    
    def __init__(self, guess_sep, guess_pa, fitboxsize, sampling_outputdir, l_only = False, fm_basename = 'Planet', null_basename = 'Null'):
        
        #Check if pymultinest is not installed and imported
        if nomultinest and v2 == False:
            raise ModuleNotFoundError('Pymultinest is not installed')
        elif nomultinest and v2 == True:
            raise ImportError('Pymultinest is not installed')
        import os
            
        # derive delta RA and delta Dec
        # in pixels
        self.guess_sep = guess_sep
        self.guess_pa = guess_pa
        
        #Set where samples get stored
        self.fm_basename = str(sampling_outputdir) + str(fm_basename) + '-'
        self.null_basename = str(sampling_outputdir) + str(null_basename) + '-'
        
        #Set which null hypothesis model to use
        self.l_only = l_only
        
        if not os.path.exists(str(sampling_outputdir)):
            os.mkdir(str(sampling_outputdir))
            
        super(PlanetEvidence, self).__init__(self.guess_sep, self.guess_pa, fitboxsize)
        
    def multifit(self):
        """
        Nested sampling parameter estimation and evidence calculation for the forward model and correlated noise.
        """
        #Copy bounds to use for priors
        sampler_bounds = np.copy(self.bounds)
        sampler_bounds[2:] = np.log(sampler_bounds[2:])
        
        def nested_prior_fm(params, ndim, nparams):
            params[0] = sampler_bounds[0][0] + params[0]*(sampler_bounds[0][1] - sampler_bounds[0][0])
            params[1] = sampler_bounds[1][0] + params[1]*(sampler_bounds[1][1] - sampler_bounds[1][0])
            params[2] = sampler_bounds[2][0] + params[2]*(sampler_bounds[2][1] - sampler_bounds[2][0])
            params[3] = sampler_bounds[3][0] + params[3]*(sampler_bounds[3][1] - sampler_bounds[3][0])
        
        def nested_prior_null3(params, ndim, nparams):
            params[0] = sampler_bounds[0][0] + params[0]*(sampler_bounds[0][1] - sampler_bounds[0][0])
            params[1] = sampler_bounds[1][0] + params[1]*(sampler_bounds[1][1] - sampler_bounds[1][0])
            params[2] = sampler_bounds[3][0] + params[2]*(sampler_bounds[3][1] - sampler_bounds[3][0])
        
        def nested_prior_null1(params, ndim, nparams):
            params[0] = sampler_bounds[3][0] + params[0]*(sampler_bounds[3][1] - sampler_bounds[3][0])
            
        global lnlike
        
        def nested_lnlike_fm(fitparams, ndim, nparams, readnoise = False):
            x_trial = fitparams[0]
            y_trial = fitparams[1]
            f_trial = fitparams[2]
            hyperparms_trial = fitparams[3]

            if readnoise:
                # last hyperparameter is a diagonal noise term. Separate it out
                readnoise_amp = np.exp(hyperparms_trial)
                hyperparms_trial = hyperparms_trial
                
            newparams = [x_trial, y_trial, f_trial, hyperparms_trial]
            return lnlike(newparams, self, self.covar)
        
        def nested_lnlike_null3(fitparams, ndim, nparams, readnoise=False):
            x_trial = fitparams[0]
            y_trial = fitparams[1]
            f_trial = -np.inf
            hyperparms_trial = fitparams[2]
            if readnoise:
                # last hyperparameter is a diagonal noise term. Separate it out
                readnoise_amp = np.exp(hyperparms_trial)
                hyperparms_trial = hyperparms_trial
                
            newparams = [x_trial, y_trial, f_trial, hyperparms_trial]

            return lnlike(newparams, self, self.covar)
        
        def nested_lnlike_null1(fitparams, ndim, nparams, readnoise=False):
            x_trial = self.guess_x
            y_trial = self.guess_y
            f_trial = -np.inf
            hyperparms_trial = fitparams[0]
            if readnoise:
                # last hyperparameter is a diagonal noise term. Separate it out
                readnoise_amp = np.exp(hyperparms_trial)
                hyperparms_trial = hyperparms_trial
                
            newparams = [x_trial, y_trial, f_trial, hyperparms_trial]

            return lnlike(newparams, self, self.covar)
        
        #Run MultiNest fir the Forward Model
        pymultinest.run(nested_lnlike_fm, nested_prior_fm, n_dims=4, outputfiles_basename=self.fm_basename, write_output=True, resume=False, init_MPI=False)
        
        #Run MultiNest for the null hypothesis
        if self.l_only == False:
            pymultinest.run(nested_lnlike_null3, nested_prior_null3, n_dims=3, outputfiles_basename=self.null_basename, write_output=True, resume=False, init_MPI=False)
        else:
            pymultinest.run(nested_lnlike_null1, nested_prior_null1, n_dims=1, outputfiles_basename=self.null_basename, write_output=True, resume=False, init_MPI=False)
    
        self.fm_data = pymultinest.Analyzer(4, outputfiles_basename=self.fm_basename)
        
        if self.l_only == False:
            self.null_data = pymultinest.Analyzer(3, outputfiles_basename=self.null_basename)
        elif self.l_only == True:
            self.null_data = pymultinest.Analyzer(1, outputfiles_basename=self.null_basename)
        
    def nested_corner_plots(self, posts, n_dim):
        import corner
        
        if n_dim == 4:
            x_data = np.ndarray.flatten(posts[:,0])
            y_data = np.ndarray.flatten(posts[:,1])
            f_data = np.ndarray.flatten(np.exp(posts[:,2]))
            hyperparam_data = np.ndarray.flatten(np.exp(posts[:,3]))
            data = np.vstack([x_data,y_data,f_data,hyperparam_data])
            all_labels = [r"x",r"y",r"$\alpha$",r"l"]
        elif n_dim == 3:
            x_data = np.ndarray.flatten(posts[:,0])
            y_data = np.ndarray.flatten(posts[:,1])
            hyperparam_data = np.ndarray.flatten(np.exp(posts[:,2]))
            data = np.vstack([x_data,y_data,hyperparam_data])
            all_labels = [r"x", r"y", r"l"]
        elif n_dim == 1:
            hyperparam_data = np.ndarray.flatten(np.exp(posts[:,0]))
            data = np.vstack([hyperparam_data])
            all_labels = [r"l"]
        fig = corner.corner(data.T, labels = all_labels, quantiles=[0.16, 0.5, 0.84])
        
        return fig
    
    def fit_plots(self): 
        fm_data = pymultinest.Analyzer(4, outputfiles_basename=self.fm_basename)
        fm_posts = fm_data.get_equal_weighted_posterior()
        fm_corner = self.nested_corner_plots(fm_posts, n_dim=4)
        
        if self.l_only == False:
            null_data = pymultinest.Analyzer(3, outputfiles_basename=self.null_basename)
            null_posts = null_data.get_equal_weighted_posterior()
            null_corner = self.nested_corner_plots(null_posts, n_dim=3)
        
        elif self.l_only == True:
            null_data = pymultinest.Analyzer(1, outputfiles_basename=self.null_basename)
            null_posts = null_data.get_equal_weighted_posterior()
            null_corner = self.nested_corner_plots(null_posts, n_dim=1)
            
        return fm_corner, null_corner
    
    def fit_stats(self):
        fm_stats = self.fm_data.get_stats()
        null_stats = self.null_data.get_stats()
        
        return fm_stats, null_stats

    def fm_residuals(self):
        fm_stats = self.fm_data.get_stats()
        self.fit_x = ParamRange(fm_stats['modes'][0]['mean'][0], fm_stats['modes'][0]['sigma'][0])
        self.fit_y = ParamRange(fm_stats['modes'][0]['mean'][1], fm_stats['modes'][0]['sigma'][1])
        self.fit_flux =  ParamRange(np.exp(fm_stats['modes'][0]['mean'][2]), np.exp(fm_stats['modes'][0]['sigma'][2]))
        self.covar_params = ParamRange(np.exp(fm_stats['modes'][0]['mean'][3]), np.exp(fm_stats['modes'][0]['sigma'][3]))
        residual_fig = self.best_fit_and_residuals()
        
         # create best fit FM
        dx = self.fit_x.bestfit - self.data_stamp_x_center
        dy = self.fit_y.bestfit - self.data_stamp_y_center

        fm_bestfit = self.fit_flux.bestfit * sinterp.shift(self.fm_stamp, [dy, dx])
        if self.padding > 0:
            fm_bestfit = fm_bestfit[self.padding:-self.padding, self.padding:-self.padding]

        # make residual map
        residual_map = self.data_stamp - fm_bestfit
        
        #Compute SNR as defined by maximum of FM best fit and the standard deviatioon of residuals
        snr_stamp =  np.max(fm_bestfit)/np.nanstd(residual_map)
        print('SNR from data stamp residuals: ' + str(snr_stamp))
        return residual_fig, snr_stamp
        
